# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import object, zip
from collections import OrderedDict

import numpy as np
import pandas as pd

from activitysim.core import chunk, config, inject, pipeline, util

logger = logging.getLogger(__name__)


def uniquify_key(dict, key, template="{} ({})"):
    """
    rename key so there are no duplicates with keys in dict

    e.g. if there is already a key named "dog", the second key will be reformatted to "dog (2)"
    """
    n = 1
    new_key = key
    while new_key in dict:
        n += 1
        new_key = template.format(key, n)

    return new_key


def evaluate_constants(expressions, constants):
    """
    Evaluate a list of constant expressions - each one can depend on the one before
    it.  These are usually used for the coefficients which have relationships
    to each other.  So ivt=.7 and then ivt_lr=ivt*.9.

    Parameters
    ----------
    expressions : Series
        the index are the names of the expressions which are
        used in subsequent evals - thus naming the expressions is required.
    constants : dict
        will be passed as the scope of eval - usually a separate set of
        constants are passed in here

    Returns
    -------
    d : dict

    """

    # FIXME why copy?
    d = {}
    for k, v in expressions.items():
        try:
            d[k] = eval(str(v), d.copy(), constants)
        except Exception as err:
            print(f"error evaluating {str(v)}")
            raise err

    return d


def read_assignment_spec(
    file_name,
    description_name="Description",
    target_name="Target",
    expression_name="Expression",
):
    """
    Read a CSV model specification into a Pandas DataFrame or Series.

    The CSV is expected to have columns for component descriptions
    targets, and expressions,

    The CSV is required to have a header with column names. For example:

        Description,Target,Expression

    Parameters
    ----------
    file_name : str
        Name of a CSV spec file.
    description_name : str, optional
        Name of the column in `fname` that contains the component description.
    target_name : str, optional
        Name of the column in `fname` that contains the component target.
    expression_name : str, optional
        Name of the column in `fname` that contains the component expression.

    Returns
    -------
    spec : pandas.DataFrame
        dataframe with three columns: ['description' 'target' 'expression']
    """

    try:
        cfg = pd.read_csv(file_name, comment="#")
    except Exception as e:
        logger.error(f"Error reading spec file: {file_name}")
        logger.error(str(e))
        raise e

    # drop null expressions
    # cfg = cfg.dropna(subset=[expression_name])

    cfg.rename(
        columns={
            target_name: "target",
            expression_name: "expression",
            description_name: "description",
        },
        inplace=True,
    )

    # backfill description
    if "description" not in cfg.columns:
        cfg.description = ""

    cfg.target = cfg.target.str.strip()
    cfg.expression = cfg.expression.str.strip()

    return cfg


class NumpyLogger(object):
    def __init__(self, logger):
        self.logger = logger
        self.target = ""
        self.expression = ""

    def write(self, msg):
        self.logger.warning(
            "numpy: %s expression: %s = %s"
            % (msg.rstrip(), str(self.target), str(self.expression))
        )


def local_utilities():
    """
    Dict of useful modules and functions to provides as locals for use in eval of expressions

    Returns
    -------
    utility_dict : dict
        name, entity pairs of locals
    """

    utility_dict = {
        "pd": pd,
        "np": np,
        "reindex": util.reindex,
        "reindex_i": util.reindex_i,
        "setting": config.setting,
        "other_than": util.other_than,
        "rng": pipeline.get_rn_generator(),
    }

    utility_dict.update(config.get_global_constants())

    return utility_dict


def is_throwaway(target):
    return target == "_"


def is_temp_scalar(target):
    return target.startswith("_") and target.isupper()


def is_temp(target):
    return target.startswith("_")


def assign_variables(
    assignment_expressions,
    df,
    locals_dict,
    df_alias=None,
    trace_rows=None,
    trace_label=None,
    chunk_log=None,
):
    """
    Evaluate a set of variable expressions from a spec in the context
    of a given data table.

    Expressions are evaluated using Python's eval function.
    Python expressions have access to variables in locals_d (and df being
    accessible as variable df.) They also have access to previously assigned
    targets as the assigned target name.

    lowercase variables starting with underscore are temp variables (e.g. _local_var)
    and not returned except in trace_results

    uppercase variables starting with underscore are temp singular variables (e.g. _LOCAL_SCALAR)
    and not returned except in trace_assigned_locals
    This is useful for defining general purpose local variables that don't vary across
    choosers or alternatives and therefore don't need to be stored as series/columns
    in the main choosers dataframe from which utilities are computed.

    Users should take care that expressions (other than temp scalar variables) should result in
    a Pandas Series (scalars will be automatically promoted to series.)

    Parameters
    ----------
    assignment_expressions : pandas.DataFrame of target assignment expressions
        target: target column names
        expression: pandas or python expression to evaluate
    df : pandas.DataFrame
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of "python" expression.
    trace_rows: series or array of bools to use as mask to select target rows to trace

    Returns
    -------
    variables : pandas.DataFrame
        Will have the index of `df` and columns named by target and containing
        the result of evaluating expression
    trace_df : pandas.DataFrame or None
        a dataframe containing the eval result values for each assignment expression
    """

    np_logger = NumpyLogger(logger)

    def is_throwaway(target):
        return target == "_"

    def is_temp_singular(target):
        return target.startswith("_") and target.isupper()

    def is_temp_series_val(target):
        return target.startswith("_")

    def to_series(x):
        if x is None or np.isscalar(x):
            return pd.Series([x] * len(df.index), index=df.index)
        return x

    assert assignment_expressions.shape[0] > 0

    trace_assigned_locals = trace_results = None
    if trace_rows is not None:
        # convert to numpy array so we can slice ndarrays as well as series

        trace_rows = np.asanyarray(trace_rows)
        if trace_rows.any():
            trace_results = OrderedDict()
            trace_assigned_locals = OrderedDict()

    # avoid touching caller's passed-in locals_d parameter (they may be looping)
    _locals_dict = local_utilities()
    if locals_dict is not None:
        _locals_dict.update(locals_dict)
    if df_alias:
        _locals_dict[df_alias] = df
    else:
        _locals_dict["df"] = df
    local_keys = list(_locals_dict.keys())

    # build a dataframe of eval results for non-temp targets
    # since we allow targets to be recycled, we want to only keep the last usage
    variables = OrderedDict()
    temps = OrderedDict()

    # need to be able to identify which variables causes an error, which keeps
    # this from being expressed more parsimoniously

    for e in zip(assignment_expressions.target, assignment_expressions.expression):
        target, expression = e

        assert isinstance(
            target, str
        ), "expected target '%s' for expression '%s' to be string not %s" % (
            target,
            expression,
            type(target),
        )

        if target in local_keys:
            logger.warning(
                "assign_variables target obscures local_d name '%s'", str(target)
            )

        if trace_label:
            logger.info(f"{trace_label}.assign_variables {target} = {expression}")

        if is_temp_singular(target) or is_throwaway(target):
            try:
                x = eval(expression, globals(), _locals_dict)
            except Exception as err:
                logger.error(
                    "assign_variables error: %s: %s", type(err).__name__, str(err)
                )
                logger.error(
                    "assign_variables expression: %s = %s", str(target), str(expression)
                )
                raise err

            if not is_throwaway(target):
                _locals_dict[target] = x
                if trace_assigned_locals is not None:
                    trace_assigned_locals[
                        uniquify_key(trace_assigned_locals, target)
                    ] = x

            continue

        try:

            # FIXME - log any numpy warnings/errors but don't raise
            np_logger.target = str(target)
            np_logger.expression = str(expression)
            saved_handler = np.seterrcall(np_logger)
            save_err = np.seterr(all="log")

            # FIXME should whitelist globals for security?
            globals_dict = {}
            expr_values = to_series(eval(expression, globals_dict, _locals_dict))

            np.seterr(**save_err)
            np.seterrcall(saved_handler)

        # except Exception as err:
        #     logger.error("assign_variables error: %s: %s", type(err).__name__, str(err))
        #     logger.error("assign_variables expression: %s = %s", str(target), str(expression))
        #     raise err

        except Exception as err:
            logger.exception(
                f"assign_variables - {type(err).__name__} ({str(err)}) evaluating: {str(expression)}"
            )
            raise err

        if not is_temp_series_val(target):
            variables[target] = expr_values

        if trace_results is not None:
            trace_results[uniquify_key(trace_results, target)] = expr_values[trace_rows]

        # just keeping track of temps so we can chunk.log_df
        if is_temp(target):
            temps[target] = expr_values
        else:
            variables[target] = expr_values

        # update locals to allows us to ref previously assigned targets
        _locals_dict[target] = expr_values

    if trace_results is not None:

        trace_results = pd.DataFrame.from_dict(trace_results)

        trace_results.index = df[trace_rows].index

        # add df columns to trace_results
        trace_results = pd.concat([df[trace_rows], trace_results], axis=1)

    assert variables, "No non-temp variables were assigned."

    if chunk_log:
        chunk.log_df(trace_label, "temps", temps)
        chunk.log_df(trace_label, "variables", variables)
        # these are going away - let caller log result df
        chunk.log_df(trace_label, "temps", None)
        chunk.log_df(trace_label, "variables", None)

    # we stored result in dict - convert to df
    variables = util.df_from_dict(variables, index=df.index)

    return variables, trace_results, trace_assigned_locals



# ActivitySim
# See full license in LICENSE.txt.

"""
reindex household_id  in households and persons tables
legacy tables have household_ids starting at 930000000
which causes headaches for activitysim's automatic generation of trip and tour ids based on hosuehold_id
(predictable trip and tour ids are used for repeatable random number stream generation)
"""

import os
import sys

import numpy as np
import pandas as pd

if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

file_names = {
    "households": "households.csv",
    "persons": "persons.csv",
    "land_use": "land_use.csv",
}

land_use_zone_col = "ZONE"
hh_zone_col = "zone_id"


def drop_and_dump(df, drop, msg, tag, output_dir):

    print("Checking for %s" % msg)
    if drop.any():
        print(
            "WARNING: dropping %s out of %s %s (%s)" % (drop.sum(), len(df), msg, tag)
        )
        df[drop].to_csv(os.path.join(output_dir, "%s.csv" % tag), index=False)
        df = df[~drop]

    return df


def create_subset(input_dir, output_dir, drop_dir):

    ###
    # land_use
    ###
    land_use_df = pd.read_csv(os.path.join(input_dir, file_names["land_use"]))
    land_use_df = land_use_df.sort_values(by=land_use_zone_col)
    land_use_df.to_csv(os.path.join(output_dir, file_names["land_use"]), index=False)

    print("zones: %s" % len(land_use_df))

    ###
    # households
    ###

    households = pd.read_csv(
        os.path.join(input_dir, file_names["households"]),
        dtype={"household_id": np.int64},
    )
    households = households.sort_values(by="household_id")
    households.rename(columns={"household_id": "legacy_household_id"}, inplace=True)

    raw_household_count = len(households)

    # all households must have a zone_id
    null_zones = households[hh_zone_col].isnull()
    households = drop_and_dump(
        households,
        null_zones,
        msg="households with null zones",
        tag="households_with_null_zones",
        output_dir=drop_dir,
    )
    households[hh_zone_col] = households[hh_zone_col].astype(np.int64)

    # all households zone_ids must be in land_use
    orphan_zones = ~households[hh_zone_col].isin(land_use_df[land_use_zone_col])
    households = drop_and_dump(
        households,
        orphan_zones,
        msg="households with unknown zones",
        tag="households_with_unknown_zones",
        output_dir=drop_dir,
    )

    # reindexed household_id as both index and column
    households.index = np.arange(1, len(households) + 1)
    households["household_id"] = households.index

    ###
    # persons
    ###
    persons = pd.read_csv(
        os.path.join(input_dir, file_names["persons"]),
        dtype={"household_id": np.int64, "person_id": np.int64},
    )
    persons = persons.sort_values(by=["household_id", "member_id"])
    persons.rename(
        columns={
            "person_id": "legacy_person_id",
            "household_id": "legacy_household_id",
        },
        inplace=True,
    )
    persons.legacy_household_id = persons.legacy_household_id.astype(np.int64)

    raw_person_count = len(persons)

    assert not persons.legacy_household_id.isnull().any()

    orphan_persons = ~persons.legacy_household_id.isin(households.legacy_household_id)
    persons = drop_and_dump(
        persons,
        orphan_persons,
        msg="persons without households",
        tag="persons_without_households",
        output_dir=drop_dir,
    )

    persons = pd.merge(
        persons,
        households[["legacy_household_id", "household_id"]],
        left_on="legacy_household_id",
        right_on="legacy_household_id",
        how="left",
    )
    assert not persons.household_id.isnull().any()
    persons.household_id = persons.household_id.astype(np.int64)

    # reindexed person_id as both index and column
    persons.index = np.arange(1, len(persons) + 1)
    persons["person_id"] = persons.index

    # check that we have the right number of persons in every household"
    assert (persons.groupby("household_id").size() == households.persons).all()

    # check that all persons in household have different member_id"
    persons_with_dupe_member_id = persons.duplicated(
        ["household_id", "member_id"], keep="first"
    )
    household_ids_with_dupe_member_id = persons.household_id[
        persons_with_dupe_member_id
    ].unique()
    households_with_dupe_members = households.household_id.isin(
        household_ids_with_dupe_member_id
    )
    persons_in_households_with_dupe_members = persons.household_id.isin(
        household_ids_with_dupe_member_id
    )

    print(
        "%s of %s persons_with_dupe_member_id"
        % (persons_with_dupe_member_id.sum(), len(persons))
    )
    persons = drop_and_dump(
        persons,
        persons_in_households_with_dupe_members,
        msg="persons in households with duplicate (household_id, member_id)",
        tag="persons_in_households_with_dupe_member_id",
        output_dir=drop_dir,
    )

    households = drop_and_dump(
        households,
        households_with_dupe_members,
        msg="households with duplicate persons.member_id",
        tag="households_with_dupe_member_id",
        output_dir=drop_dir,
    )

    missing_member1 = ~households.household_id.isin(
        persons.household_id[persons.member_id == 1]
    )
    # print("%s of %s households missing member_id 1" % (missing_member1.sum(), len(households)))
    assert not missing_member1.any()

    print(
        "Writing %s households. Dropped %s"
        % (len(households), raw_household_count - len(households))
    )
    households.to_csv(os.path.join(output_dir, file_names["households"]), index=False)

    print(
        "Writing %s persons. Dropped %s"
        % (len(persons), raw_person_count - len(persons))
    )
    persons.to_csv(os.path.join(output_dir, file_names["persons"]), index=False)


create_subset(input_dir="data_raw/", output_dir="data/", drop_dir="data_raw/dropped")



# ActivitySim
# See full license in LICENSE.txt.

import logging

import numpy as np
import openmatrix as omx
import pandas as pd

from activitysim.core import config, expressions, inject, los, pipeline

logger = logging.getLogger(__name__)


@inject.step()
def write_trip_matrices(network_los):
    """
    Write trip matrices step.

    Adds boolean columns to local trips table via annotation expressions,
    then aggregates trip counts and writes OD matrices to OMX.  Save annotated
    trips table to pipeline if desired.

    Writes taz trip tables for one and two zone system.  Writes taz and tap
    trip tables for three zone system.  Add ``is_tap:True`` to the settings file
    to identify an output matrix as tap level trips as opposed to taz level trips.

    For one zone system, uses the land use table for the set of possible tazs.
    For two zone system, uses the taz skim zone names for the set of possible tazs.
    For three zone system, uses the taz skim zone names for the set of possible tazs
    and uses the tap skim zone names for the set of possible taps.

    """

    trips = inject.get_table("trips", None)
    if trips is None:
        # this step is a NOP if there is no trips table
        # this might legitimately happen if they comment out some steps to debug but still want write_tables
        # this saves them the hassle of remembering to comment out this step
        logger.warning(
            f"write_trip_matrices returning empty-handed because there is no trips table"
        )
        return

    model_settings = config.read_model_settings("write_trip_matrices.yaml")
    trips_df = annotate_trips(trips, network_los, model_settings)

    if bool(model_settings.get("SAVE_TRIPS_TABLE")):
        pipeline.replace_table("trips", trips_df)

    if "parking_location" in config.setting("models"):
        parking_settings = config.read_model_settings("parking_location_choice.yaml")
        parking_taz_col_name = parking_settings["ALT_DEST_COL_NAME"]
        if parking_taz_col_name in trips_df:
            trips_df.loc[trips_df[parking_taz_col_name] > 0, "destination"] = trips_df[
                parking_taz_col_name
            ]
        # Also need address the return trip

    # write matrices by zone system type
    if network_los.zone_system == los.ONE_ZONE:  # taz trips written to taz matrices
        logger.info("aggregating trips one zone...")
        aggregate_trips = trips_df.groupby(["origin", "destination"], sort=False).sum()

        # use the average household weight for all trips in the origin destination pair
        hh_weight_col = model_settings.get("HH_EXPANSION_WEIGHT_COL")
        aggregate_weight = (
            trips_df[["origin", "destination", hh_weight_col]]
            .groupby(["origin", "destination"], sort=False)
            .mean()
        )
        aggregate_trips[hh_weight_col] = aggregate_weight[hh_weight_col]

        orig_vals = aggregate_trips.index.get_level_values("origin")
        dest_vals = aggregate_trips.index.get_level_values("destination")

        # use the land use table for the set of possible tazs
        zone_index = pipeline.get_table("land_use").index
        assert all(zone in zone_index for zone in orig_vals)
        assert all(zone in zone_index for zone in dest_vals)

        _, orig_index = zone_index.reindex(orig_vals)
        _, dest_index = zone_index.reindex(dest_vals)

        write_matrices(
            aggregate_trips, zone_index, orig_index, dest_index, model_settings
        )

    elif network_los.zone_system == los.TWO_ZONE:  # maz trips written to taz matrices
        logger.info("aggregating trips two zone...")
        trips_df["otaz"] = (
            pipeline.get_table("land_use").reindex(trips_df["origin"]).TAZ.tolist()
        )
        trips_df["dtaz"] = (
            pipeline.get_table("land_use").reindex(trips_df["destination"]).TAZ.tolist()
        )
        aggregate_trips = trips_df.groupby(["otaz", "dtaz"], sort=False).sum()

        # use the average household weight for all trips in the origin destination pair
        hh_weight_col = model_settings.get("HH_EXPANSION_WEIGHT_COL")
        aggregate_weight = (
            trips_df[["otaz", "dtaz", hh_weight_col]]
            .groupby(["otaz", "dtaz"], sort=False)
            .mean()
        )
        aggregate_trips[hh_weight_col] = aggregate_weight[hh_weight_col]

        orig_vals = aggregate_trips.index.get_level_values("otaz")
        dest_vals = aggregate_trips.index.get_level_values("dtaz")

        zone_index = pd.Index(network_los.get_tazs(), name="TAZ")
        assert all(zone in zone_index for zone in orig_vals)
        assert all(zone in zone_index for zone in dest_vals)

        _, orig_index = zone_index.reindex(orig_vals)
        _, dest_index = zone_index.reindex(dest_vals)

        write_matrices(
            aggregate_trips, zone_index, orig_index, dest_index, model_settings
        )

    elif (
        network_los.zone_system == los.THREE_ZONE
    ):  # maz trips written to taz and tap matrices

        logger.info("aggregating trips three zone taz...")
        trips_df["otaz"] = (
            pipeline.get_table("land_use").reindex(trips_df["origin"]).TAZ.tolist()
        )
        trips_df["dtaz"] = (
            pipeline.get_table("land_use").reindex(trips_df["destination"]).TAZ.tolist()
        )
        aggregate_trips = trips_df.groupby(["otaz", "dtaz"], sort=False).sum()

        # use the average household weight for all trips in the origin destination pair
        hh_weight_col = model_settings.get("HH_EXPANSION_WEIGHT_COL")
        aggregate_weight = (
            trips_df[["otaz", "dtaz", hh_weight_col]]
            .groupby(["otaz", "dtaz"], sort=False)
            .mean()
        )
        aggregate_trips[hh_weight_col] = aggregate_weight[hh_weight_col]

        orig_vals = aggregate_trips.index.get_level_values("otaz")
        dest_vals = aggregate_trips.index.get_level_values("dtaz")

        zone_index = pd.Index(network_los.get_tazs(), name="TAZ")
        assert all(zone in zone_index for zone in orig_vals)
        assert all(zone in zone_index for zone in dest_vals)

        _, orig_index = zone_index.reindex(orig_vals)
        _, dest_index = zone_index.reindex(dest_vals)

        write_matrices(
            aggregate_trips, zone_index, orig_index, dest_index, model_settings
        )

        logger.info("aggregating trips three zone tap...")
        aggregate_trips = trips_df.groupby(["btap", "atap"], sort=False).sum()

        # use the average household weight for all trips in the origin destination pair
        hh_weight_col = model_settings.get("HH_EXPANSION_WEIGHT_COL")
        aggregate_weight = (
            trips_df[["btap", "atap", hh_weight_col]]
            .groupby(["btap", "atap"], sort=False)
            .mean()
        )
        aggregate_trips[hh_weight_col] = aggregate_weight[hh_weight_col]

        orig_vals = aggregate_trips.index.get_level_values("btap")
        dest_vals = aggregate_trips.index.get_level_values("atap")

        zone_index = pd.Index(network_los.get_taps(), name="TAP")
        assert all(zone in zone_index for zone in orig_vals)
        assert all(zone in zone_index for zone in dest_vals)

        _, orig_index = zone_index.reindex(orig_vals)
        _, dest_index = zone_index.reindex(dest_vals)

        write_matrices(
            aggregate_trips, zone_index, orig_index, dest_index, model_settings, True
        )


def annotate_trips(trips, network_los, model_settings):
    """
    Add columns to local trips table. The annotator has
    access to the origin/destination skims and everything
    defined in the model settings CONSTANTS.

    Pipeline tables can also be accessed by listing them under
    TABLES in the preprocessor settings.
    """

    trips_df = trips.to_frame()

    trace_label = "trip_matrices"

    skim_dict = network_los.get_default_skim_dict()

    # setup skim keys
    if "trip_period" not in trips_df:
        trips_df["trip_period"] = network_los.skim_time_period_label(trips_df.depart)
    od_skim_wrapper = skim_dict.wrap("origin", "destination")
    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key="origin", dest_key="destination", dim3_key="trip_period"
    )
    skims = {"od_skims": od_skim_wrapper, "odt_skims": odt_skim_stack_wrapper}

    locals_dict = {}
    constants = config.get_model_constants(model_settings)
    if constants is not None:
        locals_dict.update(constants)

    expressions.annotate_preprocessors(
        trips_df, locals_dict, skims, model_settings, trace_label
    )

    # Data will be expanded by an expansion weight column from
    # the households pipeline table, if specified in the model settings.
    hh_weight_col = model_settings.get("HH_EXPANSION_WEIGHT_COL")

    if hh_weight_col and hh_weight_col not in trips_df:
        logger.info("adding '%s' from households to trips table" % hh_weight_col)
        household_weights = pipeline.get_table("households")[hh_weight_col]
        trips_df[hh_weight_col] = trips_df.household_id.map(household_weights)

    return trips_df


def write_matrices(
    aggregate_trips, zone_index, orig_index, dest_index, model_settings, is_tap=False
):
    """
    Write aggregated trips to OMX format.

    The MATRICES setting lists the new OMX files to write.
    Each file can contain any number of 'tables', each specified by a
    table key ('name') and a trips table column ('data_field') to use
    for aggregated counts.

    Any data type may be used for columns added in the annotation phase,
    but the table 'data_field's must be summable types: ints, floats, bools.
    """

    matrix_settings = model_settings.get("MATRICES")

    if not matrix_settings:
        logger.error("Missing MATRICES setting in write_trip_matrices.yaml")

    for matrix in matrix_settings:
        matrix_is_tap = matrix.get("is_tap", False)

        if matrix_is_tap == is_tap:  # only write tap matrices to tap matrix files
            filename = matrix.get("file_name")
            filepath = config.output_file_path(filename)
            logger.info("opening %s" % filepath)
            file = omx.open_file(filepath, "w")  # possibly overwrite existing file
            table_settings = matrix.get("tables")

            for table in table_settings:
                table_name = table.get("name")
                col = table.get("data_field")

                if col not in aggregate_trips:
                    logger.error(f"missing {col} column in aggregate_trips DataFrame")
                    return

                hh_weight_col = model_settings.get("HH_EXPANSION_WEIGHT_COL")
                if hh_weight_col:
                    aggregate_trips[col] = (
                        aggregate_trips[col] / aggregate_trips[hh_weight_col]
                    )

                data = np.zeros((len(zone_index), len(zone_index)))
                data[orig_index, dest_index] = aggregate_trips[col]
                logger.debug(
                    "writing %s sum %0.2f" % (table_name, aggregate_trips[col].sum())
                )
                file[table_name] = data  # write to file

            # include the index-to-zone map in the file
            logger.info(
                "adding %s mapping for %s zones to %s"
                % (zone_index.name, zone_index.size, filename)
            )
            file.create_mapping(zone_index.name, zone_index.to_numpy())

            logger.info("closing %s" % filepath)
            file.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing

from .util import estimation

logger = logging.getLogger(__name__)


@inject.step()
def free_parking(persons_merged, persons, chunk_size, trace_hh_id):
    """ """

    trace_label = "free_parking"
    model_settings_file_name = "free_parking.yaml"

    choosers = persons_merged.to_frame()
    choosers = choosers[choosers.workplace_zone_id > -1]
    logger.info("Running %s with %d persons", trace_label, len(choosers))

    model_settings = config.read_model_settings(model_settings_file_name)
    estimator = estimation.manager.begin_estimation("free_parking")

    constants = config.get_model_constants(model_settings)

    # - preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        locals_d = {}
        if constants is not None:
            locals_d.update(constants)

        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_d,
            trace_label=trace_label,
        )

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(choosers)

    choices = simulate.simple_simulate(
        choosers=choosers,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="free_parking_at_work",
        estimator=estimator,
    )

    free_parking_alt = model_settings["FREE_PARKING_ALT"]
    choices = choices == free_parking_alt

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "persons", "free_parking_at_work"
        )
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    persons = persons.to_frame()
    persons["free_parking_at_work"] = (
        choices.reindex(persons.index).fillna(0).astype(bool)
    )

    pipeline.replace_table("persons", persons)

    tracing.print_summary(
        "free_parking", persons.free_parking_at_work, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(persons, label=trace_label, warn_if_empty=True)



import os
import sys


def main():

    # set all these before we import numpy or any other math library
    if len(sys.argv) > 1 and sys.argv[1] == "benchmark":
        os.environ["MKL_NUM_THREADS"] = "1"
        os.environ["OMP_NUM_THREADS"] = "1"
        os.environ["OPENBLAS_NUM_THREADS"] = "1"
        os.environ["NUMBA_NUM_THREADS"] = "1"
        os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
        os.environ["NUMEXPR_NUM_THREADS"] = "1"

    from activitysim import __doc__, __version__
    from activitysim.cli import CLI, benchmark, create, run

    asim = CLI(version=__version__, description=__doc__)
    asim.add_subcommand(
        name="run",
        args_func=run.add_run_args,
        exec_func=run.run,
        description=run.run.__doc__,
    )
    asim.add_subcommand(
        name="create",
        args_func=create.add_create_args,
        exec_func=create.create,
        description=create.create.__doc__,
    )
    asim.add_subcommand(
        name="benchmark",
        args_func=benchmark.make_asv_argparser,
        exec_func=benchmark.benchmark,
        description=benchmark.benchmark.__doc__,
    )
    sys.exit(asim.execute())



import argparse
import os

import numpy as np
import openmatrix as omx
import pandas as pd

MAZ_OFFSET = 0

segments = {
    "test": (149, 215),  # SUPER_DIST_25==1, has univ
    "z500": (0, 500),
    "full": (0, 10000),
}

land_use_zone_col = "ZONE"
hh_zone_col = "zone_id"
num_full_skim_files = 2

parser = argparse.ArgumentParser(description="crop PSRC raw_data")
parser.add_argument(
    "segment_name",
    metavar="segment_name",
    type=str,
    nargs=1,
    help=f"geography segmentation (e.g. full)",
)

parser.add_argument(
    "-c",
    "--check_geography",
    default=False,
    action="store_true",
    help="check consistency of MAZ, TAZ zone_ids and foreign keys & write orphan_households file",
)

args = parser.parse_args()


segment_name = args.segment_name[0]
check_geography = args.check_geography

assert segment_name in segments.keys(), f"Unknown seg: {segment_name}"
zone_min, zone_max = segments[segment_name]

input_dir = "./data_raw"
output_dir = f"./data_{segment_name}"


print(f"check_geography {check_geography}")

if not os.path.isdir(output_dir):
    print(f"creating output directory {output_dir}")
    os.mkdir(output_dir)


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def integerize_id_columns(df, table_name):
    columns = [
        "ZONE",
        "SUPER_DIST_25",
        "zone_id",
        "household_id",
        "person_id",
        "MAZ",
        "TAZ",
    ]
    for c in df.columns:
        if c in columns:
            print(f"converting {table_name}.{c} to int")
            if df[c].isnull().any():
                print(df[c][df[c].isnull()])
            df[c] = df[c].astype(int)


def read_csv(file_name, integerize=True):
    df = pd.read_csv(input_path(file_name))

    print(f"read {file_name} {df.shape}")

    return df


def to_csv(df, file_name):
    print(f"writing {file_name} {df.shape} {output_path(file_name)}")
    df.to_csv(output_path(file_name), index=False)


if check_geography:

    # ######## check for orphan_households not in any maz in land_use
    land_use = read_csv("land_use.csv")

    households = read_csv("households.csv")
    orphan_households = households[
        ~households[hh_zone_col].isin(land_use[land_use_zone_col])
    ]
    print(f"{len(orphan_households)} orphan_households")

    if len(orphan_households) > 0:
        # write orphan_households to INPUT directory (since it doesn't belong in output)
        file_name = "orphan_households.csv"
        print(
            f"writing {file_name} {orphan_households.shape} to {input_path(file_name)}"
        )
        orphan_households.to_csv(input_path(file_name), index=False)


#
# land_use
#
land_use = read_csv("land_use.csv")
land_use = land_use[
    (land_use[land_use_zone_col] >= zone_min)
    & (land_use[land_use_zone_col] <= zone_max)
]
integerize_id_columns(land_use, "land_use")
land_use = land_use.sort_values(land_use_zone_col)

# move index col to front
land_use.insert(0, land_use_zone_col, land_use.pop(land_use_zone_col))

to_csv(land_use, "land_use.csv")

# # make sure we have some HSENROLL and COLLFTE, even for very for small samples
# if land_use['HSENROLL'].sum() == 0:
#     assert segment_name != 'full', f"land_use['HSENROLL'] is 0 for full sample!"
#     land_use['HSENROLL'] = land_use['AGE0519']
#     print(f"\nWARNING: land_use.HSENROLL is 0, so backfilled with AGE0519\n")
#
# if land_use['COLLFTE'].sum() == 0:
#     assert segment_name != 'full', f"land_use['COLLFTE'] is 0 for full sample!"
#     land_use['COLLFTE'] = land_use['HSENROLL']
#     print(f"\nWARNING: land_use.COLLFTE is 0, so backfilled with HSENROLL\n")

#
# households
#
households = read_csv("households.csv")
households = households[households[hh_zone_col].isin(land_use[land_use_zone_col])]
integerize_id_columns(households, "households")

# move index col to front
households.insert(0, "household_id", households.pop("household_id"))

to_csv(households, "households.csv")

#
# persons
#
persons = read_csv("persons.csv")
persons = persons[persons["household_id"].isin(households.household_id)]
integerize_id_columns(persons, "persons")

# move index col to front
persons.insert(0, "person_id", persons.pop("person_id"))


to_csv(persons, "persons.csv")

#
# skims
#
omx_infile_name = "skims.omx"
skim_data_type = np.float32

omx_in = omx.open_file(input_path(omx_infile_name))
print(f"omx_in shape {omx_in.shape()}")

for m in omx_in.listMappings():
    offset_map = omx_in.mapentries(m)
    # otherwise we will have to offset_map zone_indexes we use to slice skim below
    assert (offset_map == np.arange(len(offset_map)) + 1).all()
    # print(f"{m}\n{offset_map}")

zone = land_use.sort_values(land_use_zone_col)[[land_use_zone_col]]
zone.index = zone[land_use_zone_col] - 1
zone_indexes = zone.index.tolist()  # index of TAZ in skim (zero-based, no mapping)
zone_labels = zone[land_use_zone_col].tolist()  # TAZ zone_ids in omx index order

# create
num_outfiles = num_full_skim_files if segment_name == "full" else 1
if num_outfiles == 1:
    omx_out = [omx.open_file(output_path(f"skims.omx"), "w")]
else:
    omx_out = [
        omx.open_file(output_path(f"skims{i+1}.omx"), "w") for i in range(num_outfiles)
    ]

for omx_file in omx_out:
    omx_file.create_mapping(land_use_zone_col, zone_labels)

iskim = 0
for mat_name in omx_in.list_matrices():

    # make sure we have a vanilla numpy array, not a CArray
    m = np.asanyarray(omx_in[mat_name]).astype(skim_data_type)
    m = m[zone_indexes, :][:, zone_indexes]
    print(f"{mat_name} {m.shape}")

    omx_file = omx_out[iskim % num_outfiles]
    omx_file[mat_name] = m
    iskim += 1


omx_in.close()
for omx_file in omx_out:
    omx_file.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import zip
from collections import OrderedDict

import numpy as np
import pandas as pd

from . import chunk, config, logit, simulate, tracing

logger = logging.getLogger(__name__)

DUMP = False

ALT_CHOOSER_ID = "_chooser_id"


def eval_interaction_utilities(
    spec, df, locals_d, trace_label, trace_rows, estimator=None, log_alt_losers=False
):
    """
    Compute the utilities for a single-alternative spec evaluated in the context of df

    We could compute the utilities for interaction datasets just as we do for simple_simulate
    specs with multiple alternative columns by calling eval_variables and then computing the
    utilities by matrix-multiplication of eval results with the utility coefficients in the
    spec alternative columns.

    But interaction simulate computes the utilities of each alternative in the context of a
    separate row in interaction dataset df, and so there is only one alternative in spec.
    This turns out to be quite a bit faster (in this special case) than the pandas dot function.

    For efficiency, we combine eval_variables and multiplication of coefficients into a single step,
    so we don't have to create a separate column for each partial utility. Instead, we simply
    multiply the eval result by a single alternative coefficient and sum the partial utilities.


    spec : dataframe
        one row per spec expression and one col with utility coefficient

    df : dataframe
        cross join (cartesian product) of choosers with alternatives
        combines columns of choosers and alternatives
        len(df) == len(choosers) * len(alternatives)
        index values (non-unique) are index values from alternatives df

    interaction_utilities : dataframe
        the utility of each alternative is sum of the partial utilities determined by the
        various spec expressions and their corresponding coefficients
        yielding a dataframe  with len(interaction_df) rows and one utility column
        having the same index as interaction_df (non-unique values from alternatives df)

    Returns
    -------
    utilities : pandas.DataFrame
        Will have the index of `df` and a single column of utilities

    """
    trace_label = tracing.extend_trace_label(trace_label, "eval_interaction_utils")
    logger.info("Running eval_interaction_utilities on %s rows" % df.shape[0])

    with chunk.chunk_log(trace_label):

        assert len(spec.columns) == 1

        # avoid altering caller's passed-in locals_d parameter (they may be looping)
        locals_d = locals_d.copy() if locals_d is not None else {}

        # add df for startswith('@') eval expressions
        locals_d["df"] = df

        def to_series(x):
            if np.isscalar(x):
                return pd.Series([x] * len(df), index=df.index)
            if isinstance(x, np.ndarray):
                return pd.Series(x, index=df.index)
            return x

        if trace_rows is not None and trace_rows.any():
            # # convert to numpy array so we can slice ndarrays as well as series
            # trace_rows = np.asanyarray(trace_rows)
            assert type(trace_rows) == np.ndarray
            trace_eval_results = OrderedDict()
        else:
            trace_eval_results = None

        check_for_variability = config.setting("check_for_variability")

        # need to be able to identify which variables causes an error, which keeps
        # this from being expressed more parsimoniously

        utilities = pd.DataFrame({"utility": 0.0}, index=df.index)

        chunk.log_df(trace_label, "eval.utilities", utilities)

        no_variability = has_missing_vals = 0

        if estimator:
            # ensure alt_id from interaction_dataset is available in expression_values_df for
            # estimator.write_interaction_expression_values and eventual omnibus table assembly
            alt_id = estimator.get_alt_id()
            assert alt_id in df.columns
            expression_values_df = df[[alt_id]]

            # FIXME estimation_requires_chooser_id_in_df_column
            # estimation requires that chooser_id is either in index or a column of interaction_dataset
            # so it can be reformatted (melted) and indexed by chooser_id and alt_id
            # we assume caller has this under control if index is named
            # bug - location choice has df index_name zone_id but should be person_id????
            if df.index.name is None:
                chooser_id = estimator.get_chooser_id()
                assert (
                    chooser_id in df.columns
                ), "Expected to find choose_id column '%s' in interaction dataset" % (
                    chooser_id,
                )
                assert df.index.name is None
                expression_values_df[chooser_id] = df[chooser_id]

        if isinstance(spec.index, pd.MultiIndex):
            exprs = spec.index.get_level_values(simulate.SPEC_EXPRESSION_NAME)
            labels = spec.index.get_level_values(simulate.SPEC_LABEL_NAME)
        else:
            exprs = spec.index
            labels = spec.index

        for expr, label, coefficient in zip(exprs, labels, spec.iloc[:, 0]):
            try:

                # - allow temps of form _od_DIST@od_skim['DIST']
                if expr.startswith("_"):

                    target = expr[: expr.index("@")]
                    rhs = expr[expr.index("@") + 1 :]
                    v = to_series(eval(rhs, globals(), locals_d))

                    # update locals to allows us to ref previously assigned targets
                    locals_d[target] = v
                    chunk.log_df(trace_label, target, v)  # track temps stored in locals

                    if trace_eval_results is not None:
                        trace_eval_results[expr] = v[trace_rows]

                    # don't add temps to utility sums
                    # they have a non-zero dummy coefficient to avoid being removed from spec as NOPs
                    continue

                if expr.startswith("@"):
                    v = to_series(eval(expr[1:], globals(), locals_d))
                else:
                    v = df.eval(expr)

                if check_for_variability and v.std() == 0:
                    logger.info(
                        "%s: no variability (%s) in: %s"
                        % (trace_label, v.iloc[0], expr)
                    )
                    no_variability += 1

                # FIXME - how likely is this to happen? Not sure it is really a problem?
                if check_for_variability and np.count_nonzero(v.isnull().values) > 0:
                    logger.info("%s: missing values in: %s" % (trace_label, expr))
                    has_missing_vals += 1

                if estimator:
                    # in case we modified expression_values_df index
                    expression_values_df.insert(
                        loc=len(expression_values_df.columns),
                        column=label,
                        value=v.values if isinstance(v, pd.Series) else v,
                    )

                utility = (v * coefficient).astype("float")

                if log_alt_losers:

                    assert ALT_CHOOSER_ID in df
                    max_utils_by_chooser = utility.groupby(df[ALT_CHOOSER_ID]).max()

                    if (max_utils_by_chooser < simulate.ALT_LOSER_UTIL).any():

                        losers = max_utils_by_chooser[
                            max_utils_by_chooser < simulate.ALT_LOSER_UTIL
                        ]
                        logger.warning(
                            f"{trace_label} - {len(losers)} choosers of {len(max_utils_by_chooser)} "
                            f"with prohibitive utilities for all alternatives for expression: {expr}"
                        )

                        # loser_df = df[df[ALT_CHOOSER_ID].isin(losers.index)]
                        # print(f"\nloser_df\n{loser_df}\n")
                        # print(f"\nloser_max_utils_by_chooser\n{losers}\n")
                        # bug

                    del max_utils_by_chooser

                utilities.utility += utility

                if trace_eval_results is not None:

                    # expressions should have been uniquified when spec was read
                    # (though we could do it here if need be...)
                    # expr = assign.uniquify_key(trace_eval_results, expr, template="{} # ({})")
                    assert expr not in trace_eval_results

                    trace_eval_results[expr] = v[trace_rows]
                    k = "partial utility (coefficient = %s) for %s" % (
                        coefficient,
                        expr,
                    )
                    trace_eval_results[k] = v[trace_rows] * coefficient

                del v
                # chunk.log_df(trace_label, 'v', None)

            except Exception as err:
                logger.exception(
                    f"{trace_label} - {type(err).__name__} ({str(err)}) evaluating: {str(expr)}"
                )
                raise err

        if estimator:
            estimator.log(
                "eval_interaction_utilities write_interaction_expression_values %s"
                % trace_label
            )
            estimator.write_interaction_expression_values(expression_values_df)
            del expression_values_df

        if no_variability > 0:
            logger.warning(
                "%s: %s columns have no variability" % (trace_label, no_variability)
            )

        if has_missing_vals > 0:
            logger.warning(
                "%s: %s columns have missing values" % (trace_label, has_missing_vals)
            )

        if trace_eval_results is not None:
            trace_eval_results["total utility"] = utilities.utility[trace_rows]

            trace_eval_results = pd.DataFrame.from_dict(trace_eval_results)
            trace_eval_results.index = df[trace_rows].index

            # add df columns to trace_results
            trace_eval_results = pd.concat([df[trace_rows], trace_eval_results], axis=1)
            chunk.log_df(trace_label, "eval.trace_eval_results", trace_eval_results)

        chunk.log_df(trace_label, "v", None)
        chunk.log_df(trace_label, "eval.utilities", None)  # out of out hands...
        chunk.log_df(trace_label, "eval.trace_eval_results", None)

    return utilities, trace_eval_results


def _interaction_simulate(
    choosers,
    alternatives,
    spec,
    skims=None,
    locals_d=None,
    sample_size=None,
    trace_label=None,
    trace_choice_name=None,
    log_alt_losers=False,
    estimator=None,
):
    """
    Run a MNL simulation in the situation in which alternatives must
    be merged with choosers because there are interaction terms or
    because alternatives are being sampled.

    Parameters are same as for public function interaction_simulate

    spec : dataframe
        one row per spec expression and one col with utility coefficient

    interaction_df : dataframe
        cross join (cartesian product) of choosers with alternatives
        combines columns of choosers and alternatives
        len(df) == len(choosers) * len(alternatives)
        index values (non-unique) are index values from alternatives df

    interaction_utilities : dataframe
        the utility of each alternative is sum of the partial utilities determined by the
        various spec expressions and their corresponding coefficients
        yielding a dataframe  with len(interaction_df) rows and one utility column
        having the same index as interaction_df (non-unique values from alternatives df)

    utilities : dataframe
        dot product of model_design.dot(spec)
        yields utility value for element in the cross product of choosers and alternatives
        this is then reshaped as a dataframe with one row per chooser and one column per alternative

    probs : dataframe
        utilities exponentiated and converted to probabilities
        same shape as utilities, one row per chooser and one column for alternative

    positions : series
        choices among alternatives with the chosen alternative represented
        as the integer index of the selected alternative column in probs

    choices : series
        series with the alternative chosen for each chooser
        the index is same as choosers
        and the series value is the alternative df index of chosen alternative

    Returns
    -------
    ret : pandas.Series
        A series where index should match the index of the choosers DataFrame
        and values will match the index of the alternatives DataFrame -
        choices are simulated in the standard Monte Carlo fashion
    """

    trace_label = tracing.extend_trace_label(trace_label, "interaction_simulate")
    have_trace_targets = tracing.has_trace_targets(choosers)

    if have_trace_targets:
        tracing.trace_df(choosers, tracing.extend_trace_label(trace_label, "choosers"))
        tracing.trace_df(
            alternatives,
            tracing.extend_trace_label(trace_label, "alternatives"),
            slicer="NONE",
            transpose=False,
        )

    if len(spec.columns) > 1:
        raise RuntimeError("spec must have only one column")

    sample_size = sample_size or len(alternatives)

    if sample_size > len(alternatives):
        logger.debug(
            "clipping sample size %s to len(alternatives) %s"
            % (sample_size, len(alternatives))
        )
        sample_size = min(sample_size, len(alternatives))

    # if using skims, copy index into the dataframe, so it will be
    # available as the "destination" for the skims dereference below
    if skims is not None and alternatives.index.name not in alternatives:
        alternatives = alternatives.copy()
        alternatives[alternatives.index.name] = alternatives.index

    # cross join choosers and alternatives (cartesian product)
    # for every chooser, there will be a row for each alternative
    # index values (non-unique) are from alternatives df
    alt_index_id = estimator.get_alt_id() if estimator else None
    chooser_index_id = ALT_CHOOSER_ID if log_alt_losers else None

    interaction_df = logit.interaction_dataset(
        choosers,
        alternatives,
        sample_size,
        alt_index_id=alt_index_id,
        chooser_index_id=chooser_index_id,
    )
    chunk.log_df(trace_label, "interaction_df", interaction_df)

    if skims is not None:
        simulate.set_skim_wrapper_targets(interaction_df, skims)

    # evaluate expressions from the spec multiply by coefficients and sum
    # spec is df with one row per spec expression and one col with utility coefficient
    # column names of model_design match spec index values
    # utilities has utility value for element in the cross product of choosers and alternatives
    # interaction_utilities is a df with one utility column and one row per row in model_design
    if have_trace_targets:
        trace_rows, trace_ids = tracing.interaction_trace_rows(
            interaction_df, choosers, sample_size
        )

        tracing.trace_df(
            interaction_df[trace_rows],
            tracing.extend_trace_label(trace_label, "interaction_df"),
            slicer="NONE",
            transpose=False,
        )
    else:
        trace_rows = trace_ids = None
    interaction_utilities, trace_eval_results = eval_interaction_utilities(
        spec,
        interaction_df,
        locals_d,
        trace_label,
        trace_rows,
        estimator=estimator,
        log_alt_losers=log_alt_losers,
    )
    chunk.log_df(trace_label, "interaction_utilities", interaction_utilities)

    # print(f"interaction_df {interaction_df.shape}")
    # print(f"interaction_utilities {interaction_utilities.shape}")

    del interaction_df
    chunk.log_df(trace_label, "interaction_df", None)

    if have_trace_targets:
        tracing.trace_interaction_eval_results(
            trace_eval_results,
            trace_ids,
            tracing.extend_trace_label(trace_label, "eval"),
        )

        tracing.trace_df(
            interaction_utilities[trace_rows],
            tracing.extend_trace_label(trace_label, "interaction_utils"),
            slicer="NONE",
            transpose=False,
        )

    # reshape utilities (one utility column and one row per row in model_design)
    # to a dataframe with one row per chooser and one column per alternative
    utilities = pd.DataFrame(
        interaction_utilities.values.reshape(len(choosers), sample_size),
        index=choosers.index,
    )
    chunk.log_df(trace_label, "utilities", utilities)

    if have_trace_targets:
        tracing.trace_df(
            utilities,
            tracing.extend_trace_label(trace_label, "utils"),
            column_labels=["alternative", "utility"],
        )

    tracing.dump_df(DUMP, utilities, trace_label, "utilities")

    # convert to probabilities (utilities exponentiated and normalized to probs)
    # probs is same shape as utilities, one row per chooser and one column for alternative
    probs = logit.utils_to_probs(
        utilities, trace_label=trace_label, trace_choosers=choosers
    )
    chunk.log_df(trace_label, "probs", probs)

    del utilities
    chunk.log_df(trace_label, "utilities", None)

    if have_trace_targets:
        tracing.trace_df(
            probs,
            tracing.extend_trace_label(trace_label, "probs"),
            column_labels=["alternative", "probability"],
        )

    # make choices
    # positions is series with the chosen alternative represented as a column index in probs
    # which is an integer between zero and num alternatives in the alternative sample
    positions, rands = logit.make_choices(
        probs, trace_label=trace_label, trace_choosers=choosers
    )
    chunk.log_df(trace_label, "positions", positions)
    chunk.log_df(trace_label, "rands", rands)

    # need to get from an integer offset into the alternative sample to the alternative index
    # that is, we want the index value of the row that is offset by <position> rows into the
    # tranche of this choosers alternatives created by cross join of alternatives and choosers
    # offsets is the offset into model_design df of first row of chooser alternatives
    offsets = np.arange(len(positions)) * sample_size
    # resulting Int64Index has one element per chooser row and is in same order as choosers
    choices = interaction_utilities.index.take(positions + offsets)

    # create a series with index from choosers and the index of the chosen alternative
    choices = pd.Series(choices, index=choosers.index)
    chunk.log_df(trace_label, "choices", choices)

    if have_trace_targets:
        tracing.trace_df(
            choices,
            tracing.extend_trace_label(trace_label, "choices"),
            columns=[None, trace_choice_name],
        )
        tracing.trace_df(
            rands,
            tracing.extend_trace_label(trace_label, "rands"),
            columns=[None, "rand"],
        )

    return choices


def interaction_simulate(
    choosers,
    alternatives,
    spec,
    log_alt_losers=False,
    skims=None,
    locals_d=None,
    sample_size=None,
    chunk_size=0,
    trace_label=None,
    trace_choice_name=None,
    estimator=None,
):

    """
    Run a simulation in the situation in which alternatives must
    be merged with choosers because there are interaction terms or
    because alternatives are being sampled.

    optionally (if chunk_size > 0) iterates over choosers in chunk_size chunks

    Parameters
    ----------
    choosers : pandas.DataFrame
        DataFrame of choosers
    alternatives : pandas.DataFrame
        DataFrame of alternatives - will be merged with choosers, currently
        without sampling
    spec : pandas.DataFrame
        A Pandas DataFrame that gives the specification of the variables to
        compute and the coefficients for each variable.
        Variable specifications must be in the table index and the
        table should have only one column of coefficients.
    skims : Skims object
        The skims object is used to contain multiple matrices of
        origin-destination impedances.  Make sure to also add it to the
        locals_d below in order to access it in expressions.  The *only* job
        of this method in regards to skims is to call set_df with the
        dataframe that comes back from interacting choosers with
        alternatives.  See the skims module for more documentation on how
        the skims object is intended to be used.
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    sample_size : int, optional
        Sample alternatives with sample of given size.  By default is None,
        which does not sample alternatives.
    chunk_size : int
        if chunk_size > 0 iterates over choosers in chunk_size chunks
    trace_label: str
        This is the label to be used  for trace log file entries and dump file names
        when household tracing enabled. No tracing occurs if label is empty or None.
    trace_choice_name: str
        This is the column label to be used in trace file csv dump of choices

    Returns
    -------
    choices : pandas.Series
        A series where index should match the index of the choosers DataFrame
        and values will match the index of the alternatives DataFrame -
        choices are simulated in the standard Monte Carlo fashion
    """

    trace_label = tracing.extend_trace_label(trace_label, "interaction_simulate")

    assert len(choosers) > 0

    result_list = []
    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        choosers, chunk_size, trace_label
    ):

        choices = _interaction_simulate(
            chooser_chunk,
            alternatives,
            spec,
            skims=skims,
            locals_d=locals_d,
            sample_size=sample_size,
            trace_label=chunk_trace_label,
            trace_choice_name=trace_choice_name,
            log_alt_losers=log_alt_losers,
            estimator=estimator,
        )

        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

    # FIXME: this will require 2X RAM
    # if necessary, could append to hdf5 store on disk:
    # http://pandas.pydata.org/pandas-docs/stable/io.html#id2
    if len(result_list) > 1:
        choices = pd.concat(result_list)

    assert len(choices.index == len(choosers.index))

    return choices



# ActivitySim
# See full license in LICENSE.txt.
import logging
import multiprocessing
import os
import sys
import time
import traceback
from collections import OrderedDict

import numpy as np
import pandas as pd
import yaml

from activitysim.core import config, inject, mem, pipeline, tracing, util
from activitysim.core.config import setting

logger = logging.getLogger(__name__)

LAST_CHECKPOINT = "_"

MEM_TRACE_TICKS = 5

"""
mp_tasks - activitysim multiprocessing overview

Activitysim runs a list of models sequentially, performing various computational operations
on tables. Model steps can modify values in existing tables, add columns, or create additional
tables. Activitysim provides the facility, via expression files, to specify vectorized operations
on data tables. The ability to vectorize operations depends upon the independence of the
computations performed on the vectorized elements.

Python is agonizingly slow performing scalar operations sequentially on large datasets, so
vectorization (using pandas and/or numpy) is essential for good performance.

Fortunately most activity based model simulation steps are row independent at the household,
person, tour, or trip level. The decisions for one household are independent of the choices
made by other households. Thus it is (generally speaking) possible to run an entire simulation
on a household sample with only one household, and get the same result for that household as
you would running the simulation on a thousand households. (See the shared data section below
for an exception to this highly convenient situation.)

The random number generator supports this goal by providing streams of random numbers
for each households and person that are mutually independent and repeatable across model runs
and processes.

To the extent that simulation model steps are row independent, we can implement most simulations
as a series of vectorized operations on pandas DataFrames and numpy arrays. These vectorized
operations are much faster than sequential python because they are implemented by native code
(compiled C) and are to some extent multi-threaded. But the benefits of numpy multi-processing are
limited because they only apply to atomic numpy or pandas calls, and as soon as control returns
to python it is single-threaded and slow.

Multi-threading is not an attractive strategy to get around the python performance problem because
of the limitations imposed by python's global interpreter lock (GIL). Rather than struggling with
python multi-threading, this module uses the python multiprocessing to parallelize certain models.

Because of activitysim's modular and extensible architecture, we don't hardwire the multiprocessing
architecture. The specification of which models should be run in parallel, how many processers
should be used, and the segmentation of the data between processes are all specified in the
settings config file. For conceptual simplicity, the single processing model as treated as
dominant (because even though in practice multiprocessing may be the norm for production runs,
the single-processing model will be used in development and debugging and keeping it dominant
will tend to concentrate the multiprocessing-specific code in one place and prevent multiprocessing
considerations from permeating the code base obscuring the model-specific logic.

The primary function of the multiprocessing settings are to identify distinct stages of
computation, and to specify how many simultaneous processes should be used to perform them,
and how the data to be treated should be apportioned between those processes. We assume that
the data can be apportioned between subprocesses according to the index of a single primary table
(e.g. households) or else are by derivative or dependent tables that reference that table's index
(primary key) with a ref_col (foreign key) sharing the name of the primary table's key.

Generally speaking, we assume that any new tables that are created are directly dependent on the
previously existing tables, and all rows in new tables are either attributable to previously
existing rows in the pipeline tables, or are global utility tables that are identical across
sub-processes.

Note: There are a few exceptions to 'row independence', such as school and location choice models,
where the model behavior is externally constrained or adjusted. For instance, we want school
location choice to match known aggregate school enrollments by zone. Similarly, a parking model
(not yet implemented) might be constrained by availability. These situations require special
handling.

::

    models:
      ### mp_initialize step
      - initialize_landuse
      - compute_accessibility
      - initialize_households
      ### mp_households step
      - school_location
      - workplace_location
      - auto_ownership_simulate
      - free_parking
      ### mp_summarize step
      - write_tables

    multiprocess_steps:
      - name: mp_initialize
        begin: initialize_landuse
      - name: mp_households
        begin: school_location
        num_processes: 2
        slice:
          tables:
            - households
            - persons
      - name: mp_summarize
        begin: write_tables

The multiprocess_steps setting above annotates the models list to indicate that the simulation
should be broken into three steps.

The first multiprocess_step (mp_initialize) begins with the initialize_landuse step and is
implicity single-process because there is no 'slice' key indicating how to apportion the tables.
This first step includes all models listed in the 'models' setting up until the first step
in the next multiprocess_steps.

The second multiprocess_step (mp_households) starts with the school location model and continues
through auto_ownership_simulate. The 'slice' info indicates that the tables should be sliced by
households, and that persons is a dependent table and so and persons with a ref_col (foreign key
column with the same name as the Households table index) referencing a household record should be
taken to 'belong' to that household. Similarly, any other table that either share an index
(i.e. having the same name) with either the households or persons table, or have a ref_col to
either of their indexes, should also be considered a dependent table.

The num_processes setting of 2 indicates that the pipeline should be split in two, and half of the
households should be apportioned into each subprocess pipeline, and all dependent tables should
likewise be apportioned accordingly. All other tables (e.g. land_use) that do share an index (name)
or have a ref_col should be considered mirrored and be included in their entirety.

The primary table is sliced by num_processes-sized strides. (e.g. for num_processes == 2, the
sub-processes get every second record starting at offsets 0 and 1 respectively. All other dependent
tables slices are based (directly or indirectly) on this primary stride segmentation of the primary
table index.

Two separate sub-process are launched (num_processes == 2) and each passed the name of their
apportioned pipeline file. They execute independently and if they terminate successfully, their
contents are then coalesced into a single pipeline file whose tables should then be essentially
the same as it had been generated by a single process.

We assume that any new tables that are created by the sub-processes are directly dependent on the
previously primary tables or are mirrored. Thus we can coalesce the sub-process pipelines by
concatenating the primary and dependent tables and simply retaining any copy of the mirrored tables
(since they should all be identical.)

The third multiprocess_step (mp_summarize) then is handled in single-process mode and runs the
write_tables model, writing the results, but also leaving the tables in the pipeline, with
essentially the same tables and results as if the whole simulation had been run as a single process.

"""

"""
shared data

Although multiprocessing subprocesses each have their (apportioned) pipeline, they also share some
data passed to them by the parent process. There are essentially two types of shared data.

read-only shared data

Skim files are read-only and take up a lot of RAM, so we share them across sub-processes, loading
them into shared-memory (multiprocessing.sharedctypes.RawArray) in the parent process and passing
them to the child sub-processes when they are launched/forked. (unlike ordinary python data,
sharedctypes are not pickled and reconstituted, but passed through to the subprocess by address
when launch/forked multiprocessing.Process. Since they are read-only, no Locks are required to
access their data safely. The receiving process needs to know to wrap them using numpy.frombuffer
but they can thereafter be treated as ordinary numpy arrays.

read-write shared memory

There are a few circumstances in which the assumption of row independence breaks down.
This happens if the model must respect some aggregated resource or constraint such as school
enrollments or parking availability. In these cases, the individual choice models have to be
influenced or constrained in light of aggregate choices.

Currently school and workplace location choice are the only such aggregate constraints.
The details of these are handled by the shadow_pricing module (q.v.), and our only concern here
is the need to provide shared read-write data buffers for communication between processes.
It is worth noting here that the shared buffers are instances of multiprocessing.Array which
incorporates a multiprocessing.Lock object to mediate access of the underlying data. You might
think that the existence of such a lock would make shared access pretty straightforward, but
this is not the case as the level of locking is very low, reportedly not very performant, and
essentially useless in any event since we want to use numpy.frombuffer to wrap and handle them
as numpy arrays. The Lock is a convenient bundled locking primative, but shadow_pricing rolls
its own semaphore system using the Lock.

FIXME - The code below knows that it need to allocate skim and shadow price buffers by calling
the appropriate methods in abm.tables.skims and abm.tables.shadow_pricing to allocate shared
buffers. This is not very extensible and should be generalized.

"""

# FIXME - pathological knowledge of abm.tables.skims and abm.tables.shadow_pricing (see note above)


def log(msg, level, write_to_log_file=True):

    process_name = multiprocessing.current_process().name

    if not write_to_log_file:
        print(f"############ mp_tasks - {process_name} - {msg}")

    if write_to_log_file:
        with config.open_log_file("mp_tasks_log.txt", "a") as log_file:
            print(f"mp_tasks - {process_name} - {msg}", file=log_file)

    if write_to_log_file:
        # logger.log(level, f"mp_tasks - {process_name} - {msg}")
        logger.log(level, msg)


def debug(msg, write_to_log_file=True):
    log(msg, level=logging.DEBUG, write_to_log_file=write_to_log_file)


def info(msg, write_to_log_file=True):
    log(msg, level=logging.INFO, write_to_log_file=write_to_log_file)


def warning(msg, write_to_log_file=True):
    log(msg, level=logging.WARNING, write_to_log_file=write_to_log_file)


def error(msg, write_to_log_file=True):
    log(msg, level=logging.ERROR, write_to_log_file=write_to_log_file)


def exception(msg, write_to_log_file=True):

    process_name = multiprocessing.current_process().name

    if not write_to_log_file:
        print(f"mp_tasks - {process_name} - {msg}")
        print(f"---\n{traceback.format_exc()}---")

    with config.open_log_file("mp_tasks_log.txt", "a") as log_file:
        print(f"---\nmp_tasks - {process_name} - {msg}", file=log_file)
        traceback.print_exc(limit=10, file=log_file)
        print("---", file=log_file)

    if write_to_log_file:
        logger.log(logging.ERROR, f"mp_tasks - {process_name} - {msg}")
        logger.log(logging.ERROR, f"\n---\n{traceback.format_exc()}---\n")


"""
### child process methods (called within sub process)
"""


def pipeline_table_keys(pipeline_store):
    """
    return dict of current (as of last checkpoint) pipeline tables
    and their checkpoint-specific hdf5_keys

    This facilitates reading pipeline tables directly from a 'raw' open pandas.HDFStore without
    opening it as a pipeline (e.g. when apportioning and coalescing pipelines)

    We currently only ever need to do this from the last checkpoint, so the ability to specify
    checkpoint_name is not required, and thus omitted.

    Parameters
    ----------
    pipeline_store : open hdf5 pipeline_store

    Returns
    -------
    checkpoint_name : name of the checkpoint
    checkpoint_tables : dict {<table_name>: <table_key>}

    """

    checkpoints = pipeline_store[pipeline.CHECKPOINT_TABLE_NAME]

    # don't currently need this capability...
    # if checkpoint_name:
    #     # specified checkpoint row as series
    #     i = checkpoints[checkpoints[pipeline.CHECKPOINT_NAME] == checkpoint_name].index[0]
    #     checkpoint = checkpoints.loc[i]
    # else:

    # last checkpoint row as series
    checkpoint = checkpoints.iloc[-1]
    checkpoint_name = checkpoint.loc[pipeline.CHECKPOINT_NAME]

    # series with table name as index and checkpoint_name as value
    checkpoint_tables = checkpoint[~checkpoint.index.isin(pipeline.NON_TABLE_COLUMNS)]

    # omit dropped tables with empty checkpoint name
    checkpoint_tables = checkpoint_tables[checkpoint_tables != ""]

    # hdf5 key is <table_name>/<checkpoint_name>
    checkpoint_tables = {
        table_name: pipeline.pipeline_table_key(table_name, checkpoint_name)
        for table_name, checkpoint_name in checkpoint_tables.items()
    }

    # checkpoint name and series mapping table name to hdf5 key for tables in that checkpoint
    return checkpoint_name, checkpoint_tables


def build_slice_rules(slice_info, pipeline_tables):
    """
    based on slice_info for current step from run_list, generate a recipe for slicing
    the tables in the pipeline (passed in tables parameter)

    slice_info is a dict with two well-known keys:
        'tables': required list of table names (order matters!)
        'except': optional list of tables not to slice even if they have a sliceable index name

    Note: tables listed in slice_info must appear in same order and before any others in tables dict

    The index of the first table in the 'tables' list is the primary_slicer.

    Any other tables listed are dependent tables with either ref_cols to the primary_slicer
    or with the same index (i.e. having an index with the same name). This cascades, so any
    tables dependent on the primary_table can in turn have dependent tables that will be sliced
    by index or ref_col.

    For instance, if the primary_slicer is households, then persons can be sliced because it
    has a ref_col to (column with the same same name as) the household table index. And the
    tours table can be sliced since it has a ref_col to persons. Tables can also be sliced
    by index. For instance the person_windows table can be sliced because it has an index with
    the same names as the persons table.

    slice_info from multiprocess_steps

    ::

        slice:
          tables:
            - households
            - persons

    tables from pipeline

    +-----------------+--------------+---------------+
    | Table Name      | Index        | ref_col       |
    +=================+==============+===============+
    | households      | household_id |               |
    +-----------------+--------------+---------------+
    | persons         | person_id    | household_id  |
    +-----------------+--------------+---------------+
    | person_windows  | person_id    |               |
    +-----------------+--------------+---------------+
    | accessibility   | zone_id      |               |
    +-----------------+--------------+---------------+

    generated slice_rules dict

    ::

        households:
           slice_by: primary       <- primary table is sliced in num_processors-sized strides
        persons:
           source: households
           slice_by: column
           column:  household_id   <- slice by ref_col (foreign key) to households
        person_windows:
           source: persons
           slice_by: index         <- slice by index of persons table
        accessibility:
           slice_by:               <- mirrored (non-dependent) tables don't get sliced
        land_use:
           slice_by:


    Parameters
    ----------
    slice_info : dict
        'slice' info from run_list for this step

    pipeline_tables : dict {<table_name>, <pandas.DataFrame>}
        dict of all tables from the pipeline keyed by table name

    Returns
    -------
    slice_rules : dict
    """

    slicer_table_names = slice_info["tables"]
    slicer_table_exceptions = slice_info.get("except", [])
    primary_slicer = slicer_table_names[0]

    # - ensure that tables listed in slice_info appear in correct order and before any others
    tables = OrderedDict([(table_name, None) for table_name in slicer_table_names])

    for table_name in pipeline_tables.keys():
        tables[table_name] = pipeline_tables[table_name]

    if primary_slicer not in tables:
        raise RuntimeError("primary slice table '%s' not in pipeline" % primary_slicer)

    # allow wildcard 'True' to avoid slicing (or coalescing) any tables no explicitly listed in slice_info.tables
    # populationsim uses slice.except wildcards to avoid listing control tables (etc) that should not be sliced,
    # followed by a slice.coalesce directive to explicitly list the omnibus tables created by the subprocesses.
    # So don't change this behavior withoyt testing populationsim multiprocess!
    if slicer_table_exceptions is True:
        debug(
            f"slice.except wildcard (True): excluding all tables not explicitly listed in slice.tables"
        )
        slicer_table_exceptions = [t for t in tables if t not in slicer_table_names]

    # dict mapping slicer table_name to index name
    # (also presumed to be name of ref col name in referencing table)
    slicer_ref_cols = OrderedDict()

    if slicer_table_exceptions == "*":
        slicer_table_exceptions = [t for t in tables if t not in slicer_table_names]

    # build slice rules for loaded tables
    slice_rules = OrderedDict()
    for table_name, df in tables.items():

        rule = {}
        if table_name == primary_slicer:
            # slice primary apportion table
            rule = {"slice_by": "primary"}
        elif table_name in slicer_table_exceptions:
            rule["slice_by"] = None
        else:
            for slicer_table_name in slicer_ref_cols:
                if df.index.name is not None and (
                    df.index.name == tables[slicer_table_name].index.name
                ):
                    # slice df with same index name as a known slicer
                    rule = {"slice_by": "index", "source": slicer_table_name}
                else:
                    # if df has a column with same name as the ref_col (index) of a slicer?
                    try:
                        source, ref_col = next(
                            (t, c)
                            for t, c in slicer_ref_cols.items()
                            if c in df.columns
                        )
                        # then we can use that table to slice this df
                        rule = {
                            "slice_by": "column",
                            "column": ref_col,
                            "source": source,
                        }
                    except StopIteration:
                        rule["slice_by"] = None

        if rule["slice_by"]:
            # cascade sliceability
            slicer_ref_cols[table_name] = df.index.name

        slice_rules[table_name] = rule

    for table_name, rule in slice_rules.items():
        if rule["slice_by"] is not None:
            debug(
                f"### table_name: {table_name} slice_rules: {slice_rules[table_name]}"
            )
    debug(f"### slicer_ref_cols: {slicer_ref_cols}")

    return slice_rules


def apportion_pipeline(sub_proc_names, step_info):
    """
    apportion pipeline for multiprocessing step

    create pipeline files for sub_procs, apportioning data based on slice_rules

    Called at the beginning of a multiprocess step prior to launching the sub-processes
    Pipeline files have well known names (pipeline file name prefixed by subjob name)

    Parameters
    ----------
    sub_proc_names : list of str
        names of the sub processes to apportion
    step_info : dict
        step_info from multiprocess_steps for step we are apportioning pipeline tables for

    Returns
    -------
    creates apportioned pipeline files for each sub job
    """

    slice_info = step_info.get("slice", None)
    multiprocess_step_name = step_info.get("name", None)

    pipeline_file_name = inject.get_injectable("pipeline_file_name")

    # ensure that if we are resuming, we don't apportion any tables from future model steps
    last_checkpoint_in_previous_multiprocess_step = step_info.get(
        "last_checkpoint_in_previous_multiprocess_step", None
    )
    assert last_checkpoint_in_previous_multiprocess_step is not None
    pipeline.open_pipeline(resume_after=last_checkpoint_in_previous_multiprocess_step)

    # ensure all tables are in the pipeline
    checkpointed_tables = pipeline.checkpointed_tables()
    for table_name in slice_info["tables"]:
        if table_name not in checkpointed_tables:
            raise RuntimeError(f"slicer table {table_name} not found in pipeline")

    checkpoints_df = pipeline.get_checkpoints()

    # for the subprocess pipelines, keep only the last row of checkpoints and patch the last checkpoint name
    checkpoints_df = checkpoints_df.tail(1).copy()

    # load all tables from pipeline
    checkpoint_name = multiprocess_step_name
    tables = {}
    for table_name in checkpointed_tables:
        # patch last checkpoint name for all tables
        checkpoints_df[table_name] = checkpoint_name
        # load the dataframe
        tables[table_name] = pipeline.get_table(table_name)

        debug(f"loaded table {table_name} {tables[table_name].shape}")

    pipeline.close_pipeline()

    # should only be one checkpoint (named <multiprocess_step_name>)
    assert len(checkpoints_df) == 1

    # - build slice rules for loaded tables
    slice_rules = build_slice_rules(slice_info, tables)

    # - allocate sliced tables for each sub_proc
    num_sub_procs = len(sub_proc_names)
    for i in range(num_sub_procs):

        # use well-known pipeline file name
        process_name = sub_proc_names[i]
        pipeline_path = config.build_output_file_path(
            pipeline_file_name, use_prefix=process_name
        )

        # remove existing file
        try:
            os.unlink(pipeline_path)
        except OSError:
            pass

        with pd.HDFStore(pipeline_path, mode="a") as pipeline_store:

            # remember sliced_tables so we can cascade slicing to other tables
            sliced_tables = {}

            # - for each table in pipeline
            for table_name, rule in slice_rules.items():

                df = tables[table_name]

                if rule["slice_by"] is not None and num_sub_procs > len(df):

                    # almost certainly a configuration error
                    raise RuntimeError(
                        f"apportion_pipeline: multiprocess step {multiprocess_step_name} "
                        f"slice table {table_name} has fewer rows {df.shape} "
                        f"than num_processes ({num_sub_procs})."
                    )

                if rule["slice_by"] == "primary":
                    # slice primary apportion table by num_sub_procs strides
                    # this hopefully yields a more random distribution
                    # (e.g.) households are ordered by size in input store
                    # we are assuming that the primary table index is unique
                    # otherwise we should slice by strides in df.index.unique
                    # we could easily work around this, but it seems likely this was an error on the user's part
                    assert not df.index.duplicated().any()

                    primary_df = df[
                        np.asanyarray(list(range(df.shape[0]))) % num_sub_procs == i
                    ]
                    sliced_tables[table_name] = primary_df
                elif rule["slice_by"] == "index":
                    # slice a table with same index name as a known slicer
                    source_df = sliced_tables[rule["source"]]
                    sliced_tables[table_name] = df.loc[source_df.index]
                elif rule["slice_by"] == "column":
                    # slice a table with a recognized slicer_column
                    source_df = sliced_tables[rule["source"]]
                    sliced_tables[table_name] = df[
                        df[rule["column"]].isin(source_df.index)
                    ]
                elif rule["slice_by"] is None:
                    # don't slice mirrored tables
                    sliced_tables[table_name] = df
                else:
                    raise RuntimeError(
                        "Unrecognized slice rule '%s' for table %s"
                        % (rule["slice_by"], table_name)
                    )

                # - write table to pipeline
                hdf5_key = pipeline.pipeline_table_key(table_name, checkpoint_name)
                pipeline_store[hdf5_key] = sliced_tables[table_name]

            debug(
                f"writing checkpoints ({checkpoints_df.shape}) "
                f"to {pipeline.CHECKPOINT_TABLE_NAME} in {pipeline_path}"
            )
            pipeline_store[pipeline.CHECKPOINT_TABLE_NAME] = checkpoints_df


def coalesce_pipelines(sub_proc_names, slice_info):
    """
    Coalesce the data in the sub_processes apportioned pipelines back into a single pipeline

    We use slice_rules to distinguish sliced (apportioned) tables from mirrored tables.

    Sliced tables are concatenated to create a single omnibus table with data from all sub_procs
    but mirrored tables are the same across all sub_procs, so we can grab a copy from any pipeline.

    Parameters
    ----------
    sub_proc_names : list of str
    slice_info : dict
        slice_info from multiprocess_steps

    Returns
    -------
    creates an omnibus pipeline with coalesced data from individual sub_proc pipelines
    """

    pipeline_file_name = inject.get_injectable("pipeline_file_name")

    debug(f"coalesce_pipelines to: {pipeline_file_name}")

    # - read all tables from first process pipeline
    # FIXME - note: assumes any new tables will be present in ALL subprocess pipelines
    tables = {}
    pipeline_path = config.build_output_file_path(
        pipeline_file_name, use_prefix=sub_proc_names[0]
    )

    with pd.HDFStore(pipeline_path, mode="r") as pipeline_store:

        # hdf5_keys is a dict mapping table_name to pipeline hdf5_key
        checkpoint_name, hdf5_keys = pipeline_table_keys(pipeline_store)

        for table_name, hdf5_key in hdf5_keys.items():
            debug(f"loading table {table_name} {hdf5_key}")
            tables[table_name] = pipeline_store[hdf5_key]

    # slice.coalesce is an override  list of omnibus tables created by subprocesses that should be coalesced,
    # whether or not they satisfy the slice rules. Ordinarily all tables qualify for slicing by the slice rules
    # will be coalesced, including any new tables created by the subprocess that have sliceable indexes or ref_cols.
    # Any other new tables that don't match the slice rules will be considered mirrored. This is usually the desired
    # behavior, especially in activitysim abm models. However, if the "slice.except: True" wildcard is used, it
    # prevents the inference for newly generated tables, and this directive permits explicit specification of
    # which new tables to coalesce. Populationsim uses this wildcard except directives to avoid having to list
    # many slice exceptions, and just lists weigh tables to coalesce. So don't change this behavior without testing
    # populationsim multiprocessing!
    coalesce_tables = slice_info.get("coalesce", [])

    # report absence of any slice_info.coalesce tables not in pipeline
    # we don't require their presence in case there are tracing tables that will only be present if tracing is enabled
    for table_name in coalesce_tables:
        if table_name not in tables:
            logger.warning(
                "slicer coalesce.table %s not found in pipeline" % table_name
            )

    # - use slice rules followed by apportion_pipeline to identify mirrored tables
    # (tables that are identical in every pipeline and so don't need to be concatenated)
    slice_rules = build_slice_rules(slice_info, tables)

    # table is mirrored if no slice rule or explicitly listed in slice_info.coalesce setting
    mirrored_table_names = [
        t
        for t, rule in slice_rules.items()
        if rule["slice_by"] is None and t not in coalesce_tables
    ]
    mirrored_tables = {t: tables[t] for t in mirrored_table_names}
    omnibus_keys = {t: k for t, k in hdf5_keys.items() if t not in mirrored_table_names}

    debug(f"coalesce_pipelines to: {pipeline_file_name}")
    debug(f"mirrored_table_names: {mirrored_table_names}")
    debug(f"omnibus_keys: {omnibus_keys}")

    # assemble lists of omnibus tables from all sub_processes
    omnibus_tables = {table_name: [] for table_name in omnibus_keys}
    for process_name in sub_proc_names:
        pipeline_path = config.build_output_file_path(
            pipeline_file_name, use_prefix=process_name
        )
        logger.info(f"coalesce pipeline {pipeline_path}")

        with pd.HDFStore(pipeline_path, mode="r") as pipeline_store:
            for table_name, hdf5_key in omnibus_keys.items():
                omnibus_tables[table_name].append(pipeline_store[hdf5_key])

    # open pipeline, preserving existing checkpoints (so resume_after will work for prior steps)
    pipeline.open_pipeline("_")

    # - add mirrored tables to pipeline
    for table_name in mirrored_tables:
        df = mirrored_tables[table_name]
        info(f"adding mirrored table {table_name} {df.shape}")
        pipeline.replace_table(table_name, df)

    # - concatenate omnibus tables and add them to pipeline
    for table_name in omnibus_tables:
        df = pd.concat(omnibus_tables[table_name], sort=False)
        info(f"adding omnibus table {table_name} {df.shape}")
        pipeline.replace_table(table_name, df)

    pipeline.add_checkpoint(checkpoint_name)

    pipeline.close_pipeline()


def setup_injectables_and_logging(injectables, locutor=True):
    """
    Setup injectables (passed by parent process) within sub process

    we sometimes want only one of the sub-processes to perform an action (e.g. write shadow prices)
    the locutor flag indicates that this sub process is the designated singleton spokesperson

    Parameters
    ----------
    injectables : dict {<injectable_name>: <value>}
        dict of injectables passed by parent process
    locutor : bool
        is this sub process the designated spokesperson

    Returns
    -------
    injects injectables
    """

    # register abm steps and other abm-specific injectables
    # by default, assume we are running activitysim.abm
    # other callers (e.g. piopulationsim) will have to arrange to register their own steps and injectables
    # (presumably) in a custom run_simulation.py instead of using the 'activitysim run' command
    if not inject.is_injectable("preload_injectables"):
        from activitysim import (  # register abm steps and other abm-specific injectables
            abm,
        )

    try:

        for k, v in injectables.items():
            inject.add_injectable(k, v)

        inject.add_injectable("is_sub_task", True)
        inject.add_injectable("locutor", locutor)

        config.filter_warnings()

        process_name = multiprocessing.current_process().name
        inject.add_injectable("log_file_prefix", process_name)

    except Exception as e:
        exception(
            f"{type(e).__name__} exception while setting up injectables: {str(e)}",
            write_to_log_file=False,
        )
        raise e

    try:
        tracing.config_logger()
    except Exception as e:
        exception(f"{type(e).__name__} exception while configuring logger: {str(e)}")
        raise e


def adjust_chunk_size_for_shared_memory(chunk_size, data_buffers, num_processes):

    # even if there is only one subprocess,
    # we are separate from parent who allocated the shared memory
    # so we still need to compensate for it

    if chunk_size == 0:
        return chunk_size

    shared_memory_size = mem.shared_memory_size(data_buffers)

    if shared_memory_size == 0:
        return chunk_size

    fair_share_of_shared_memory = int(shared_memory_size / num_processes)

    adjusted_chunk_size = chunk_size - fair_share_of_shared_memory

    logger.info(
        f"adjust_chunk_size_for_shared_memory "
        f"adjusted_chunk_size {util.INT(adjusted_chunk_size)} "
        f"chunk_size {util.INT(chunk_size)} "
        f"shared_memory_size {util.INT(shared_memory_size)} "
        f"num_processes {num_processes} "
        f"fair_share_of_shared_memory {util.INT(fair_share_of_shared_memory)} "
    )

    if adjusted_chunk_size <= 0:
        raise RuntimeError(
            f"adjust_chunk_size_for_shared_memory: chunk_size too small for shared memory.  "
            f"adjusted_chunk_size: {adjusted_chunk_size}"
        )

    return adjusted_chunk_size


def run_simulation(queue, step_info, resume_after, shared_data_buffer):
    """
    run step models as subtask

    called once to run each individual sub process in multiprocess step

    Unless actually resuming resuming, resume_after will be None for first step,
    and then FINAL for subsequent steps so pipelines opened to resume where previous step left off

    Parameters
    ----------
    queue : multiprocessing.Queue
    step_info : dict
        step_info for current step from multiprocess_steps
    resume_after : str or None
    shared_data_buffer : dict
        dict of shared data (e.g. skims and shadow_pricing)
    """

    # step_label = step_info['name']

    models = step_info["models"]
    chunk_size = step_info["chunk_size"]
    num_processes = step_info["num_processes"]

    chunk_size = adjust_chunk_size_for_shared_memory(
        chunk_size, shared_data_buffer, num_processes
    )

    inject.add_injectable("data_buffers", shared_data_buffer)
    inject.add_injectable("chunk_size", chunk_size)
    inject.add_injectable("num_processes", num_processes)

    if resume_after:
        info(f"resume_after {resume_after}")

        # if they specified a resume_after model, check to make sure it is checkpointed
        if (
            resume_after != LAST_CHECKPOINT
            and resume_after
            not in pipeline.get_checkpoints()[pipeline.CHECKPOINT_NAME].values
        ):
            # if not checkpointed, then fall back to last checkpoint
            info(f"resume_after checkpoint '{resume_after}' not in pipeline.")
            resume_after = LAST_CHECKPOINT

    pipeline.open_pipeline(resume_after)
    last_checkpoint = pipeline.last_checkpoint()

    if last_checkpoint in models:
        info(f"Resuming model run list after {last_checkpoint}")
        models = models[models.index(last_checkpoint) + 1 :]

    assert inject.get_injectable("preload_injectables", None)

    t0 = tracing.print_elapsed_time()
    for model in models:

        t1 = tracing.print_elapsed_time()

        try:
            pipeline.run_model(model)
        except Exception as e:
            warning(f"{type(e).__name__} exception running {model} model: {str(e)}")
            raise e

        tracing.log_runtime(model_name=model, start_time=t1)
        queue.put({"model": model, "time": time.time() - t1})

    tracing.print_elapsed_time("run (%s models)" % len(models), t0)

    # add checkpoint with final tables even if not intermediate checkpointing
    checkpoint_name = step_info["name"]
    pipeline.add_checkpoint(checkpoint_name)

    pipeline.close_pipeline()


"""
### multiprocessing sub-process entry points
"""


def mp_run_simulation(locutor, queue, injectables, step_info, resume_after, **kwargs):
    """
    mp entry point for run_simulation

    Parameters
    ----------
    locutor
    queue
    injectables
    step_info
    resume_after : bool
    kwargs : dict
        shared_data_buffers passed as kwargs to avoid picking dict
    """

    setup_injectables_and_logging(injectables, locutor=locutor)

    debug(
        f"mp_run_simulation {step_info['name']} locutor={inject.get_injectable('locutor', False)} "
    )

    try:

        if step_info["num_processes"] > 1:
            pipeline_prefix = multiprocessing.current_process().name
            logger.debug(f"injecting pipeline_file_prefix '{pipeline_prefix}'")
            inject.add_injectable("pipeline_file_prefix", pipeline_prefix)

        shared_data_buffer = kwargs
        run_simulation(queue, step_info, resume_after, shared_data_buffer)

        mem.log_global_hwm()  # subprocess

    except Exception as e:
        exception(f"{type(e).__name__} exception caught in mp_run_simulation: {str(e)}")
        raise e


def mp_apportion_pipeline(injectables, sub_proc_names, step_info):
    """
    mp entry point for apportion_pipeline

    Parameters
    ----------
    injectables : dict
        injectables from parent
    sub_proc_names : list of str
        names of the sub processes to apportion
    step_info : dict
        step_info for multiprocess_step we are apportioning
    """

    setup_injectables_and_logging(injectables)

    try:
        apportion_pipeline(sub_proc_names, step_info)
    except Exception as e:
        exception(
            f"{type(e).__name__} exception caught in mp_apportion_pipeline: {str(e)}"
        )
        raise e


def mp_setup_skims(injectables, **kwargs):
    """
    Sub process to load skim data into shared_data

    There is no particular necessity to perform this in a sub process instead of the parent
    except to ensure that this heavyweight task has no side-effects (e.g. loading injectables)

    Parameters
    ----------
    injectables : dict
        injectables from parent
    kwargs : dict
        shared_data_buffers passed as kwargs to avoid picking dict
    """

    setup_injectables_and_logging(injectables)

    info("mp_setup_skims")

    try:
        shared_data_buffer = kwargs

        network_los_preload = inject.get_injectable("network_los_preload", None)

        if network_los_preload is not None:
            network_los_preload.load_shared_data(shared_data_buffer)

    except Exception as e:
        exception(f"{type(e).__name__} exception caught in mp_setup_skims: {str(e)}")
        raise e


def mp_coalesce_pipelines(injectables, sub_proc_names, slice_info):
    """
    mp entry point for coalesce_pipeline

    Parameters
    ----------
    injectables : dict
        injectables from parent
    sub_proc_names : list of str
        names of the sub processes to apportion
    slice_info : dict
        slice_info from multiprocess_steps
    """

    setup_injectables_and_logging(injectables)

    try:
        coalesce_pipelines(sub_proc_names, slice_info)
    except Exception as e:
        exception(
            f"{type(e).__name__} exception caught in coalesce_pipelines: {str(e)}"
        )
        raise e


"""
### main (parent) process methods
"""


def allocate_shared_skim_buffers():
    """
    This is called by the main process to allocate shared memory buffer to share with subprocs

    Note: Buffers must be allocated BEFORE network_los.load_data

    Returns
    -------
    skim_buffers : dict {<skim_tag>: <multiprocessing.RawArray>}

    """

    info("allocate_shared_skim_buffer")

    network_los = inject.get_injectable("network_los_preload", None)
    if network_los is not None:
        skim_buffers = network_los.allocate_shared_skim_buffers()
    else:
        skim_buffers = {}

    return skim_buffers


def allocate_shared_shadow_pricing_buffers():
    """
    This is called by the main process to allocate memory buffer to share with subprocs

    Returns
    -------
        multiprocessing.RawArray
    """

    info("allocate_shared_shadow_pricing_buffers")

    shadow_pricing_info = inject.get_injectable("shadow_pricing_info", None)

    if shadow_pricing_info is not None:
        from activitysim.abm.tables import shadow_pricing

        shadow_pricing_buffers = shadow_pricing.buffers_for_shadow_pricing(
            shadow_pricing_info
        )
    else:
        shadow_pricing_buffers = {}

    return shadow_pricing_buffers


def run_sub_simulations(
    injectables,
    shared_data_buffers,
    step_info,
    process_names,
    resume_after,
    previously_completed,
    fail_fast,
):
    """
    Launch sub processes to run models in step according to specification in step_info.

    If resume_after is LAST_CHECKPOINT, then pick up where previous run left off, using breadcrumbs
    from previous run. If some sub-processes completed in the prior run, then skip rerunning them.

    If resume_after specifies a checkpiont, skip checkpoints that precede the resume_after

    Drop 'completed' breadcrumbs for this run as sub-processes terminate

    Wait for all sub-processes to terminate and return list of those that completed successfully.

    Parameters
    ----------
    injectables : dict
        values to inject in subprocesses
    shared_data_buffers : dict
        dict of shared_data for sub-processes (e.g. skim and shadow pricing data)
    step_info : dict
        step_info from run_list
    process_names : list of str
        list of sub process names to in parallel
    resume_after : str or None
        name of simulation to resume after, or LAST_CHECKPOINT to resume where previous run left off
    previously_completed : list of str
        names of processes that successfully completed in previous run
    fail_fast : bool
        whether to raise error if a sub process terminates with nonzero exitcode

    Returns
    -------
    completed : list of str
        names of sub_processes that completed successfully

    """

    def log_queued_messages():
        for process, queue in zip(procs, queues):
            while not queue.empty():
                msg = queue.get(block=False)
                model_name = msg["model"]
                info(
                    f"{process.name} {model_name} : {tracing.format_elapsed_time(msg['time'])}"
                )
                mem.trace_memory_info(f"{process.name}.{model_name}.completed")

    def check_proc_status():
        # we want to drop 'completed' breadcrumb when it happens, lest we terminate
        # if fail_fast flag is set raise
        for p in procs:
            if p.exitcode is None:
                pass  # still running
            elif p.exitcode == 0:
                # completed successfully
                if p.name not in completed:
                    info(f"process {p.name} completed")
                    completed.add(p.name)
                    drop_breadcrumb(step_name, "completed", list(completed))
                    mem.trace_memory_info(f"{p.name}.completed")
            else:
                # process failed
                if p.name not in failed:
                    warning(f"process {p.name} failed with exitcode {p.exitcode}")
                    failed.add(p.name)
                    mem.trace_memory_info(f"{p.name}.failed")
                    if fail_fast:
                        warning(f"fail_fast terminating remaining running processes")
                        for op in procs:
                            if op.exitcode is None:
                                try:
                                    info(f"terminating process {op.name}")
                                    op.terminate()
                                except Exception as e:
                                    info(f"error terminating process {op.name}: {e}")
                        raise RuntimeError("Process %s failed" % (p.name,))

    step_name = step_info["name"]

    t0 = tracing.print_elapsed_time()
    info(f"run_sub_simulations step {step_name} models resume_after {resume_after}")

    # if resuming and some processes completed successfully in previous run
    if previously_completed:
        assert resume_after is not None
        assert set(previously_completed).issubset(set(process_names))

        if resume_after == LAST_CHECKPOINT:
            # if we are resuming where previous run left off, then we can skip running
            # any subprocudures that successfully complete the previous run
            process_names = [
                name for name in process_names if name not in previously_completed
            ]
            info(
                f"step {step_name}: skipping {len(previously_completed)} previously completed subprocedures"
            )
        else:
            # if we are resuming after a specific model, then force all subprocesses to run
            # (assuming if they specified a model, they really want everything after that to run)
            previously_completed = []

    # if not the first step, resume_after the last checkpoint from the previous step
    if resume_after is None and step_info["step_num"] > 0:
        resume_after = LAST_CHECKPOINT

    num_simulations = len(process_names)
    procs = []
    queues = []

    completed = set(previously_completed)
    failed = set([])  # so we can log process failure first time it happens
    drop_breadcrumb(step_name, "completed", list(completed))

    for i, process_name in enumerate(process_names):
        q = multiprocessing.Queue()
        locutor = i == 0

        args = OrderedDict(
            locutor=locutor,
            queue=q,
            injectables=injectables,
            step_info=step_info,
            resume_after=resume_after,
        )

        # debug(f"create_process {process_name} target={mp_run_simulation}")
        # for k in args:
        #     debug(f"create_process {process_name} arg {k}={args[k]}")
        # for k in shared_data_buffers:
        #     debug(f"create_process {process_name} shared_data_buffers {k}={shared_data_buffers[k]}")

        p = multiprocessing.Process(
            target=mp_run_simulation,
            name=process_name,
            args=(
                locutor,
                q,
                injectables,
                step_info,
                resume_after,
            ),
            kwargs=shared_data_buffers,
        )

        procs.append(p)
        queues.append(q)

    # - start processes
    for i, p in zip(list(range(num_simulations)), procs):
        info(f"start process {p.name}")
        p.start()

        """
        windows mmap does not handle multiple simultaneous calls from different processes for the same tagname.
        Process start causes a call to mmap to initialize the wrapper for the anonymous shared memory arrays
        in the shared_data_buffers kwargs. some of the processses fail with WinError 1450 (or similar error)
        OSError: [WinError 1450] Insufficient system resources exist to complete the requested service.
        Judging by the commented-out assert, this (or a related) issue may have been around in some form for a while.

        def __setstate__(self, state):
            self.size, self.name = self._state = state
            # Reopen existing mmap
            self.buffer = mmap.mmap(-1, self.size, tagname=self.name)
            # XXX Temporarily preventing buildbot failures while determining
            # XXX the correct long-term fix. See issue 23060
            #assert _winapi.GetLastError() == _winapi.ERROR_ALREADY_EXISTS
        """
        if sys.platform == "win32":
            time.sleep(1)

        mem.trace_memory_info(f"{p.name}.start")

    while multiprocessing.active_children():
        # log queued messages as they are received
        log_queued_messages()
        # monitor sub process status and drop breadcrumbs or fail_fast as they terminate
        check_proc_status()
        # monitor memory usage
        mem.trace_memory_info(
            "run_sub_simulations.idle", trace_ticks=mem.MEM_PARENT_TRACE_TICK_LEN
        )
        time.sleep(1)

    # clean up any messages or breadcrumbs that occurred while we slept
    log_queued_messages()
    check_proc_status()

    # no need to join() explicitly since multiprocessing.active_children joins completed procs

    for p in procs:
        assert p.exitcode is not None
        if p.exitcode:
            error(f"Process %s failed with exitcode {p.exitcode}")
            assert p.name in failed
        else:
            info(f"Process {p.name} completed with exitcode {p.exitcode}")
            assert p.name in completed

    t0 = tracing.print_elapsed_time("run_sub_simulations step %s" % step_name, t0)

    return list(completed)


def run_sub_task(p):
    """
    Run process p synchroneously,

    Return when sub process terminates, or raise error if exitcode is nonzero

    Parameters
    ----------
    p : multiprocessing.Process
    """
    info(f"#run_model running sub_process {p.name}")

    mem.trace_memory_info(f"{p.name}.start")

    t0 = tracing.print_elapsed_time()
    p.start()

    while multiprocessing.active_children():
        mem.trace_memory_info(
            "run_sub_simulations.idle", trace_ticks=mem.MEM_PARENT_TRACE_TICK_LEN
        )
        time.sleep(1)

    # no need to join explicitly since multiprocessing.active_children joins completed procs
    # p.join()

    t0 = tracing.print_elapsed_time("#run_model sub_process %s" % p.name, t0)
    # info(f'{p.name}.exitcode = {p.exitcode}')

    mem.trace_memory_info(f"run_model {p.name} completed")

    if p.exitcode:
        error(f"Process {p.name} returned exitcode {p.exitcode}")
        raise RuntimeError("Process %s returned exitcode %s" % (p.name, p.exitcode))


def drop_breadcrumb(step_name, crumb, value=True):
    """
    Add (crumb: value) to specified step in breadcrumbs and flush breadcrumbs to file
    run can be resumed with resume_after

    Breadcrumbs provides a record of steps that have been run for use when resuming
    Basically, we want to know which steps have been run, which phases completed
    (i.e. apportion, simulate, coalesce). For multi-processed simulate steps, we also
    want to know which sub-processes completed successfully, because if resume_after
    is LAST_CHECKPOINT we don't have to rerun the successful ones.

    Parameters
    ----------
    step_name : str
    crumb : str
    value : yaml-writable value

    Returns
    -------

    """
    breadcrumbs = inject.get_injectable("breadcrumbs", OrderedDict())
    breadcrumbs.setdefault(step_name, {"name": step_name})[crumb] = value
    inject.add_injectable("breadcrumbs", breadcrumbs)
    write_breadcrumbs(breadcrumbs)


def run_multiprocess(injectables):
    """
    run the steps in run_list, possibly resuming after checkpoint specified by resume_after

    we never open the pipeline since that is all done within multi-processing steps -
    mp_apportion_pipeline, run_sub_simulations, mp_coalesce_pipelines -
    each of which opens the pipeline/s and closes it/them within the sub-process
    This 'feature' makes the pipeline state a bit opaque to us, for better or worse...

    Steps may be either single or multi process.
    For multi-process steps, we need to apportion pipelines before running sub processes
    and coalesce them afterwards

    injectables arg allows propagation of setting values that were overridden on the command line
    (parent process command line arguments are not available to sub-processes in Windows)

    * allocate shared data buffers for skims and shadow_pricing
    * load shared skim data from OMX files
    * run each (single or multiprocess) step in turn

    Drop breadcrumbs along the way to facilitate resuming in a later run

    Parameters
    ----------
    run_list : dict
        annotated run_list  (including prior run breadcrumbs if resuming)
    injectables : dict
        dict of values to inject in sub-processes
    """

    mem.trace_memory_info("run_multiprocess.start")

    run_list = get_run_list()

    if not run_list["multiprocess"]:
        raise RuntimeError(
            "run_multiprocess called but multiprocess flag is %s"
            % run_list["multiprocess"]
        )

    old_breadcrumbs = run_list.get("breadcrumbs", {})

    # raise error if any sub-process fails without waiting for others to complete
    fail_fast = setting("fail_fast")
    info(f"run_multiprocess fail_fast: {fail_fast}")

    def skip_phase(phase):
        skip = old_breadcrumbs and old_breadcrumbs.get(step_name, {}).get(phase, False)
        if skip:
            info(f"Skipping {step_name} {phase}")
        return skip

    def find_breadcrumb(crumb, default=None):
        return old_breadcrumbs.get(step_name, {}).get(crumb, default)

    # - allocate shared data
    shared_data_buffers = {}

    mem.trace_memory_info("allocate_shared_skim_buffer.before")

    t0 = tracing.print_elapsed_time()
    shared_data_buffers.update(allocate_shared_skim_buffers())
    t0 = tracing.print_elapsed_time("allocate shared skim buffer", t0)
    mem.trace_memory_info("allocate_shared_skim_buffer.completed")

    # combine shared_skim_buffer and shared_shadow_pricing_buffer in shared_data_buffer
    t0 = tracing.print_elapsed_time()
    shared_data_buffers.update(allocate_shared_shadow_pricing_buffers())
    t0 = tracing.print_elapsed_time("allocate shared shadow_pricing buffer", t0)
    mem.trace_memory_info("allocate_shared_shadow_pricing_buffers.completed")

    # - mp_setup_skims
    if len(shared_data_buffers) > 0:
        run_sub_task(
            multiprocessing.Process(
                target=mp_setup_skims,
                name="mp_setup_skims",
                args=(injectables,),
                kwargs=shared_data_buffers,
            )
        )
        t0 = tracing.print_elapsed_time("setup shared_data_buffers", t0)
        mem.trace_memory_info("mp_setup_skims.completed")

    # - for each step in run list
    for step_info in run_list["multiprocess_steps"]:

        step_name = step_info["name"]

        num_processes = step_info["num_processes"]
        slice_info = step_info.get("slice", None)

        if num_processes == 1:
            sub_proc_names = [step_name]
        else:
            sub_proc_names = ["%s_%s" % (step_name, i) for i in range(num_processes)]

        # - mp_apportion_pipeline
        if not skip_phase("apportion") and num_processes > 1:
            run_sub_task(
                multiprocessing.Process(
                    target=mp_apportion_pipeline,
                    name="%s_apportion" % step_name,
                    args=(injectables, sub_proc_names, step_info),
                )
            )
        drop_breadcrumb(step_name, "apportion")

        # - run_sub_simulations
        if not skip_phase("simulate"):
            resume_after = step_info.get("resume_after", None)

            previously_completed = find_breadcrumb("completed", default=[])

            completed = run_sub_simulations(
                injectables,
                shared_data_buffers,
                step_info,
                sub_proc_names,
                resume_after,
                previously_completed,
                fail_fast,
            )

            if len(completed) != num_processes:
                raise RuntimeError(
                    "%s processes failed in step %s"
                    % (num_processes - len(completed), step_name)
                )
        drop_breadcrumb(step_name, "simulate")

        # - mp_coalesce_pipelines
        if not skip_phase("coalesce") and num_processes > 1:
            run_sub_task(
                multiprocessing.Process(
                    target=mp_coalesce_pipelines,
                    name="%s_coalesce" % step_name,
                    args=(injectables, sub_proc_names, slice_info),
                )
            )
        drop_breadcrumb(step_name, "coalesce")

    # add checkpoint with final tables even if not intermediate checkpointing
    if not pipeline.intermediate_checkpoint():
        pipeline.open_pipeline("_")
        pipeline.add_checkpoint(pipeline.FINAL_CHECKPOINT_NAME)
        pipeline.close_pipeline()

    mem.log_global_hwm()  # main process


def get_breadcrumbs(run_list):
    """
    Read, validate, and annotate breadcrumb file from previous run

    if resume_after specifies a model name, we need to determine which step it falls within,
    drop any subsequent steps, and set the 'simulate' and 'coalesce' to None so

    Extract from breadcrumbs file showing completed mp_households step with 2 processes:
    ::

        - apportion: true
          completed: [mp_households_0, mp_households_1]
          name: mp_households
          simulate: true
          coalesce: true


    Parameters
    ----------
    run_list : dict
        validated and annotated run_list from settings

    Returns
    -------
    breadcrumbs : dict
        validated and annotated breadcrumbs file from previous run
    """

    resume_after = run_list["resume_after"]
    assert resume_after is not None

    # - read breadcrumbs file from previous run
    breadcrumbs = read_breadcrumbs()

    # - can't resume multiprocess without breadcrumbs file
    if not breadcrumbs:
        error(f"empty breadcrumbs for resume_after '{resume_after}'")
        raise RuntimeError("empty breadcrumbs for resume_after '%s'" % resume_after)

    # if resume_after is specified by name
    if resume_after != LAST_CHECKPOINT:

        # breadcrumbs for steps from previous run
        previous_steps = list(breadcrumbs.keys())

        # find the run_list step resume_after is in
        resume_step = next(
            (
                step
                for step in run_list["multiprocess_steps"]
                if resume_after in step["models"]
            ),
            None,
        )

        resume_step_name = resume_step["name"]

        if resume_step_name not in previous_steps:
            error(f"resume_after model '{resume_after}' not in breadcrumbs")
            raise RuntimeError(
                "resume_after model '%s' not in breadcrumbs" % resume_after
            )

        # drop any previous_breadcrumbs steps after resume_step
        for step in previous_steps[previous_steps.index(resume_step_name) + 1 :]:
            del breadcrumbs[step]

        # if resume_after is not the last model in the step
        # then we need to rerun the simulations in that step, even if they succeeded
        if resume_after in resume_step["models"][:-1]:
            if "simulate" in breadcrumbs[resume_step_name]:
                breadcrumbs[resume_step_name]["simulate"] = None
            if "coalesce" in breadcrumbs[resume_step_name]:
                breadcrumbs[resume_step_name]["coalesce"] = None

    multiprocess_step_names = [step["name"] for step in run_list["multiprocess_steps"]]
    if list(breadcrumbs.keys()) != multiprocess_step_names[: len(breadcrumbs)]:
        raise RuntimeError(
            "last run steps don't match run list: %s" % list(breadcrumbs.keys())
        )

    return breadcrumbs


def get_run_list():
    """
    validate and annotate run_list from settings

    Assign defaults to missing settings (e.g. chunk_size)
    Build individual step model lists based on step starts
    If resuming, read breadcrumbs file for info on previous run execution status

    # annotated run_list with two steps, the second with 2 processors

    ::

        resume_after: None
        multiprocess: True
        models:
          -  initialize_landuse
          -  compute_accessibility
          -  initialize_households
          -  school_location
          -  workplace_location

        multiprocess_steps:
          step: mp_initialize
            begin: initialize_landuse
            name: mp_initialize
            models:
               - initialize_landuse
               - compute_accessibility
               - initialize_households
            num_processes: 1
            chunk_size: 0
            step_num: 0
          step: mp_households
            begin: school_location
            slice: {'tables': ['households', 'persons']}
            name: mp_households
            models:
               - school_location
               - workplace_location
            num_processes: 2
            chunk_size: 10000
            step_num: 1

    Returns
    -------
    run_list : dict
        validated and annotated run_list
    """

    models = setting("models", [])
    multiprocess_steps = setting("multiprocess_steps", [])

    resume_after = inject.get_injectable("resume_after", None) or setting(
        "resume_after", None
    )
    multiprocess = inject.get_injectable("multiprocess", False) or setting(
        "multiprocess", False
    )

    # default settings that can be overridden by settings in individual steps
    global_chunk_size = setting("chunk_size", 0) or 0
    default_mp_processes = setting("num_processes", 0) or int(
        1 + multiprocessing.cpu_count() / 2.0
    )

    if multiprocess and multiprocessing.cpu_count() == 1:
        warning("Can't multiprocess because there is only 1 cpu")

    run_list = {
        "models": models,
        "resume_after": resume_after,
        "multiprocess": multiprocess,
        # 'multiprocess_steps': multiprocess_steps  # add this later if multiprocess
    }

    if not models or not isinstance(models, list):
        raise RuntimeError("No models list in settings file")
    if resume_after == models[-1]:
        raise RuntimeError(
            "resume_after '%s' is last model in models list" % resume_after
        )

    if multiprocess:

        if not multiprocess_steps:
            raise RuntimeError(
                "multiprocess setting is %s but no multiprocess_steps setting"
                % multiprocess
            )

        # check step name, num_processes, chunk_size and presence of slice info
        num_steps = len(multiprocess_steps)
        step_names = set()
        for istep in range(num_steps):
            step = multiprocess_steps[istep]

            step["step_num"] = istep

            # - validate step name
            name = step.get("name", None)
            if not name:
                raise RuntimeError(
                    "missing name for step %s" " in multiprocess_steps" % istep
                )
            if name in step_names:
                raise RuntimeError(
                    "duplicate step name %s" " in multiprocess_steps" % name
                )
            if name in models:
                raise RuntimeError(
                    f"multiprocess_steps step name '{name}' cannot also be a model name"
                )

            step_names.add(name)

            # - validate num_processes and assign default
            num_processes = step.get("num_processes", 0)

            if not isinstance(num_processes, int) or num_processes < 0:
                raise RuntimeError(
                    "bad value (%s) for num_processes for step %s"
                    " in multiprocess_steps" % (num_processes, name)
                )

            if "slice" in step:
                if num_processes == 0:
                    info(f"Setting num_processes = {num_processes} for step {name}")
                    num_processes = default_mp_processes
                if num_processes > multiprocessing.cpu_count():
                    warning(
                        f"num_processes setting ({num_processes}) "
                        f"greater than cpu count ({ multiprocessing.cpu_count()})"
                    )
            else:
                if num_processes == 0:
                    num_processes = 1
                if num_processes > 1:
                    raise RuntimeError(
                        "num_processes > 1 but no slice info for step %s"
                        " in multiprocess_steps" % name
                    )

            multiprocess_steps[istep]["num_processes"] = num_processes

            # - validate chunk_size and assign default
            chunk_size = step.get("chunk_size", None)
            if chunk_size is None:
                if global_chunk_size > 0 and num_processes > 1:
                    chunk_size = int(round(global_chunk_size / num_processes))
                    chunk_size = max(chunk_size, 1)
                else:
                    chunk_size = global_chunk_size

            multiprocess_steps[istep]["chunk_size"] = chunk_size

        # - determine index in models list of step starts
        start_tag = "begin"
        starts = [0] * len(multiprocess_steps)
        for istep in range(num_steps):
            step = multiprocess_steps[istep]

            name = step["name"]

            slice = step.get("slice", None)
            if slice:
                if "tables" not in slice:
                    raise RuntimeError(
                        "missing tables list for step %s"
                        " in multiprocess_steps" % istep
                    )

            start = step.get(start_tag, None)
            if not name:
                raise RuntimeError(
                    "missing %s tag for step '%s' (%s)"
                    " in multiprocess_steps" % (start_tag, name, istep)
                )
            if start not in models:
                raise RuntimeError(
                    "%s tag '%s' for step '%s' (%s) not in models list"
                    % (start_tag, start, name, istep)
                )

            starts[istep] = models.index(start)

            if istep == 0 and starts[istep] != 0:
                raise RuntimeError(
                    "%s tag '%s' for first step '%s' (%s)"
                    " is not first model in models list"
                    % (start_tag, start, name, istep)
                )

            if istep > 0 and starts[istep] <= starts[istep - 1]:
                raise RuntimeError(
                    "%s tag '%s' for step '%s' (%s)"
                    " falls before that of prior step in models list"
                    % (start_tag, start, name, istep)
                )

            # remember there should always be a final checkpoint with same name as multiprocess_step name
            multiprocess_steps[istep][
                "last_checkpoint_in_previous_multiprocess_step"
            ] = (multiprocess_steps[istep - 1].get("name") if istep > 0 else None)

        # - build individual step model lists based on starts
        starts.append(len(models))  # so last step gets remaining models in list
        for istep in range(num_steps):
            step_models = models[starts[istep] : starts[istep + 1]]

            if step_models[-1][0] == LAST_CHECKPOINT:
                raise RuntimeError(
                    "Final model '%s' in step %s models list not checkpointed"
                    % (step_models[-1], name)
                )

            multiprocess_steps[istep]["models"] = step_models

        run_list["multiprocess_steps"] = multiprocess_steps

        # - add resume breadcrumbs
        if resume_after:
            breadcrumbs = get_breadcrumbs(run_list)
            if breadcrumbs:
                run_list["breadcrumbs"] = breadcrumbs

                # - add resume_after to last step
                if resume_after is not None:
                    # get_breadcrumbs should have deleted breadcrumbs for any subsequent steps
                    istep = len(breadcrumbs) - 1
                    assert (
                        resume_after == LAST_CHECKPOINT
                        or resume_after in multiprocess_steps[istep]["models"]
                    )
                    multiprocess_steps[istep]["resume_after"] = resume_after

    # - write run list to output dir
    # use log_file_path so we use (optional) log subdir and prefix process name
    with config.open_log_file("run_list.txt", "w") as f:
        print_run_list(run_list, f)

    return run_list


def print_run_list(run_list, output_file=None):
    """
    Print run_list to stdout or file (informational - not read back in)

    Parameters
    ----------
    run_list : dict
    output_file : open file
    """

    if output_file is None:
        output_file = sys.stdout

    print("resume_after:", run_list["resume_after"], file=output_file)
    print("multiprocess:", run_list["multiprocess"], file=output_file)

    print("models:", file=output_file)
    for m in run_list["models"]:
        print("  - ", m, file=output_file)

    # - print multiprocess_steps
    if run_list["multiprocess"]:
        print("\nmultiprocess_steps:", file=output_file)
        for step in run_list["multiprocess_steps"]:
            print("  step:", step["name"], file=output_file)
            for k in step:
                if isinstance(step[k], list):
                    print("    %s:" % k, file=output_file)
                    for v in step[k]:
                        print("       -", v, file=output_file)
                else:
                    print("    %s: %s" % (k, step[k]), file=output_file)

    # - print breadcrumbs
    breadcrumbs = run_list.get("breadcrumbs")
    if breadcrumbs:
        print("\nbreadcrumbs:", file=output_file)
        for step_name in breadcrumbs:
            step = breadcrumbs[step_name]
            print("  step:", step_name, file=output_file)
            for k in step:
                if isinstance(k, str):
                    print("    ", k, step[k], file=output_file)
                else:
                    print("    ", k, file=output_file)
                    for v in step[k]:
                        print("      ", v, file=output_file)


def breadcrumbs_file_path():
    # return path to breadcrumbs file in output_dir
    return config.build_output_file_path("breadcrumbs.yaml")


def read_breadcrumbs():
    """
    Read breadcrumbs file from previous run

    write_breadcrumbs wrote OrderedDict steps as list so ordered is preserved
    (step names are duplicated in steps)

    Returns
    -------
    breadcrumbs : OrderedDict
    """
    file_path = breadcrumbs_file_path()
    if not os.path.exists(file_path):
        raise IOError("Could not find saved breadcrumbs file '%s'" % file_path)
    with open(file_path, "r") as f:
        breadcrumbs = yaml.load(f, Loader=yaml.SafeLoader)
    # convert array to ordered dict keyed by step name
    breadcrumbs = OrderedDict([(step["name"], step) for step in breadcrumbs])
    return breadcrumbs


def write_breadcrumbs(breadcrumbs):
    """
    Write breadcrumbs file with execution history of multiprocess run

    Write steps as array so order is preserved (step names are duplicated in steps)

    Extract from breadcrumbs file showing completed mp_households step with 2 processes:
    ::

        - apportion: true
          coalesce: true
          completed: [mp_households_0, mp_households_1]
          name: mp_households
          simulate: true

    Parameters
    ----------
    breadcrumbs : OrderedDict
    """
    with open(breadcrumbs_file_path(), "w") as f:
        # write ordered dict as array
        breadcrumbs = [step for step in list(breadcrumbs.values())]
        yaml.dump(breadcrumbs, f)


def if_sub_task(if_is, if_isnt):
    """
    select one of two values depending whether current process is primary process or subtask

    This is primarily intended for use in yaml files to select between (e.g.) logging levels
    so main log file can display only warnings and errors from subtasks

    In yaml file, it can be used like this:

    level: !!python/object/apply:activitysim.core.mp_tasks.if_sub_task [WARNING, NOTSET]


    Parameters
    ----------
    if_is : (any type) value to return if process is a subtask
    if_isnt : (any type) value to return if process is not a subtask

    Returns
    -------
    (any type) (one of parameters if_is or if_isnt)
    """

    return if_is if inject.get_injectable("is_sub_task", False) else if_isnt



# crop marin tvpb example data processing to one county
# Ben Stabler, ben.stabler@rsginc.com, 09/17/20
# jeff doyle added code to introduce MAZ_OFFSET to avoid confusion (and detect associated errors) between zone types

import os

import openmatrix as omx
import pandas as pd

# counties = ["Marin"
# counties = ["San Francisco"
counties = ["Marin", "San Francisco"]

input_dir = "./data_3_marin"
output_dir = "./data_3_marin/crop"
MAZ_OFFSET = 100000


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def patch_maz(df, maz_offset):
    for c in df.columns:
        if c in ["MAZ", "OMAZ", "DMAZ", "mgra", "orig_mgra", "dest_mgra"]:
            df[c] += maz_offset
    return df


def read_csv(file_name):
    df = pd.read_csv(input_path(file_name))
    if MAZ_OFFSET:
        df = patch_maz(df, MAZ_OFFSET)
        print(f"\n\n{file_name}\n{df}")
    return df


def to_csv(df, file_name):
    df.to_csv(output_path(file_name), index=False)


#########
# mazs = read_csv("maz_data_asim.csv")
# taps = read_csv("tap_data.csv")
#
# print(f"max maz {mazs.MAZ.max()}")
# print(f"num maz {len(mazs.MAZ.unique())}")
# print(f"num taz {len(mazs.TAZ.unique())}")
# print(f"num tap {len(taps.TAP.unique())}")
#
# num maz 5952
# num taz 4735
# num tap 6216
#########


# 0 - get county zones

mazs = read_csv("maz_data_asim.csv")

mazs = mazs[mazs["CountyName"].isin(counties)]
to_csv(mazs, "maz_data_asim.csv")

maz_taz = mazs[["MAZ", "TAZ"]]
to_csv(mazs, "maz_taz.csv")

tazs = mazs["TAZ"].unique()
tazs.sort()
tazs_indexes = (tazs - 1).tolist()

taps = read_csv("tap_data.csv")
taps = taps[["TAP", "TAZ"]].sort_values(by="TAP")
taps = taps[taps["TAZ"].isin(tazs)]
to_csv(taps, "tap_data.csv")

# 1-based tap_ids
taps_indexes = (taps["TAP"] - 1).tolist()


# 2 - maz to tap walk, bike

maz_tap_walk = read_csv("maz_tap_walk.csv")
maz_maz_walk = read_csv("maz_maz_walk.csv")
maz_maz_bike = read_csv("maz_maz_bike.csv")

maz_tap_walk = maz_tap_walk[
    maz_tap_walk["MAZ"].isin(mazs["MAZ"]) & maz_tap_walk["TAP"].isin(taps["TAP"])
]
maz_maz_walk = maz_maz_walk[
    maz_maz_walk["OMAZ"].isin(mazs["MAZ"]) & maz_maz_walk["DMAZ"].isin(mazs["MAZ"])
]
maz_maz_bike = maz_maz_bike[
    maz_maz_bike["OMAZ"].isin(mazs["MAZ"]) & maz_maz_bike["DMAZ"].isin(mazs["MAZ"])
]

to_csv(maz_tap_walk, "maz_tap_walk.csv")
to_csv(maz_maz_walk, "maz_maz_walk.csv")
to_csv(maz_maz_bike, "maz_maz_bike.csv")


tap_lines = read_csv("tap_lines.csv")
tap_lines = tap_lines[tap_lines["TAP"].isin(taps["TAP"])]
to_csv(tap_lines, "tap_lines.csv")

# taz to tap drive data

taz_tap_drive = read_csv("maz_taz_tap_drive.csv")
taz_tap_drive = taz_tap_drive[
    taz_tap_drive["MAZ"].isin(mazs["MAZ"]) & taz_tap_drive["TAP"].isin(taps["TAP"])
]
to_csv(taz_tap_drive, "maz_taz_tap_drive.csv")


# 3 - accessibility data

access = read_csv("access.csv")
access = access[access["mgra"].isin(mazs["MAZ"])]
to_csv(access, "access.csv")


# households

households = read_csv("households_asim.csv")
households = households[households["MAZ"].isin(mazs["MAZ"])]
to_csv(households, "households_asim.csv")

# persons

persons = read_csv("persons_asim.csv")
persons = persons[persons["HHID"].isin(households["HHID"])]
to_csv(persons, "persons_asim.csv")

# tours file

work_tours = read_csv("work_tours.csv")
work_tours = work_tours[work_tours["hh_id"].isin(households["HHID"])]
work_tours = work_tours[
    work_tours["orig_mgra"].isin(mazs["MAZ"])
    & work_tours["dest_mgra"].isin(mazs["MAZ"])
]
to_csv(work_tours, "work_tours.csv")

# skims

time_periods = ["AM", "EA", "EV", "MD", "PM"]
for tp in time_periods:
    omx_file_name = "HWYSKM" + tp + "_taz_rename.omx"
    taz_file = omx.open_file(input_path(omx_file_name))
    taz_file_rename = omx.open_file(output_path(omx_file_name), "w")
    taz_file_rename.create_mapping("ZONE", tazs.tolist())
    for mat_name in taz_file.list_matrices():
        taz_file_rename[mat_name] = taz_file[mat_name][tazs_indexes, :][:, tazs_indexes]
        print(mat_name)
    taz_file.close()
    taz_file_rename.close()

for tp in time_periods:
    for skim_set in ["SET1", "SET2", "SET3"]:
        omx_file_name = "transit_skims_" + tp + "_" + skim_set + "_rename.omx"
        tap_file = omx.open_file(input_path(omx_file_name))
        tap_file_rename = omx.open_file(output_path(omx_file_name), "w")
        tap_file_rename.create_mapping("ZONE", taps["TAP"].tolist())
        for mat_name in tap_file.list_matrices():
            tap_file_rename[mat_name] = tap_file[mat_name][taps_indexes, :][
                :, taps_indexes
            ]
            print(mat_name)
        tap_file.close()
        tap_file_rename.close()



# ActivitySim
# See full license in LICENSE.txt.

import itertools
import logging
import os

import numpy as np
import pandas as pd

from activitysim.core import (
    assign,
    config,
    expressions,
    inject,
    logit,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.interaction_simulate import interaction_simulate
from activitysim.core.util import assign_in_place

from .util import estimation
from .util.mode import mode_choice_simulate

logger = logging.getLogger(__name__)


def annotate_vehicle_allocation(model_settings, trace_label):
    """
    Add columns to the tours table in the pipeline according to spec.

    Parameters
    ----------
    model_settings : dict
    trace_label : str
    """
    tours = inject.get_table("tours").to_frame()
    expressions.assign_columns(
        df=tours,
        model_settings=model_settings.get("annotate_tours"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_tours"),
    )
    pipeline.replace_table("tours", tours)


def get_skim_dict(network_los, choosers):
    """
    Returns a dictionary of skim wrappers to use in expression writing.

    Skims have origin as home_zone_id and destination as the tour destination.

    Parameters
    ----------
    network_los : activitysim.core.los.Network_LOS object
    choosers : pd.DataFrame

    Returns
    -------
    skims : dict
        index is skim wrapper name, value is the skim wrapper
    """
    skim_dict = network_los.get_default_skim_dict()
    orig_col_name = "home_zone_id"
    dest_col_name = "destination"

    out_time_col_name = "start"
    in_time_col_name = "end"
    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="out_period"
    )
    dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="in_period"
    )

    choosers["in_period"] = network_los.skim_time_period_label(
        choosers[in_time_col_name]
    )
    choosers["out_period"] = network_los.skim_time_period_label(
        choosers[out_time_col_name]
    )

    skims = {
        "odt_skims": odt_skim_stack_wrapper.set_df(choosers),
        "dot_skims": dot_skim_stack_wrapper.set_df(choosers),
    }
    return skims


@inject.step()
def vehicle_allocation(
    persons,
    households,
    vehicles,
    tours,
    tours_merged,
    network_los,
    chunk_size,
    trace_hh_id,
):
    """Selects a vehicle for each occupancy level for each tour.

    Alternatives consist of the up to the number of household vehicles plus one
    option for non-household vehicles.

    The model will be run once for each tour occupancy defined in the model yaml.
    Output tour table will columns added for each occupancy level.

    The user may also augment the `tours` tables with new vehicle
    type-based fields specified via the annotate_tours option.

    Parameters
    ----------
    persons : orca.DataFrameWrapper
    households : orca.DataFrameWrapper
    vehicles : orca.DataFrameWrapper
    vehicles_merged : orca.DataFrameWrapper
    tours : orca.DataFrameWrapper
    tours_merged : orca.DataFrameWrapper
    chunk_size : orca.injectable
    trace_hh_id : orca.injectable
    """
    trace_label = "vehicle_allocation"
    model_settings_file_name = "vehicle_allocation.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get("MODE_CHOICE_LOGSUM_COLUMN_NAME")

    estimator = estimation.manager.begin_estimation("vehicle_allocation")

    model_spec_raw = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec_raw, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    locals_dict = {}
    locals_dict.update(constants)
    locals_dict.update(coefficients_df)

    # ------ constructing alternatives from model spec and joining to choosers
    vehicles_wide = vehicles.to_frame().pivot_table(
        index="household_id",
        columns="vehicle_num",
        values="vehicle_type",
        aggfunc=lambda x: "".join(x),
    )

    alts_from_spec = model_spec.columns
    # renaming vehicle numbers to alternative names in spec
    vehicle_alt_columns_dict = {}
    for veh_num in range(1, len(alts_from_spec)):
        vehicle_alt_columns_dict[veh_num] = alts_from_spec[veh_num - 1]
    vehicles_wide.rename(columns=vehicle_alt_columns_dict, inplace=True)

    # if the number of vehicles is less than the alternatives, fill with NA
    # e.g. all households only have 1 or 2 vehicles because of small sample size,
    #   still need columns for alternatives 3 and 4
    for veh_num, col_name in vehicle_alt_columns_dict.items():
        if col_name not in vehicles_wide.columns:
            vehicles_wide[col_name] = ""

    # last entry in spec is the non-hh-veh option
    assert (
        alts_from_spec[-1] == "non_hh_veh"
    ), "Last option in spec needs to be non_hh_veh"
    vehicles_wide[alts_from_spec[-1]] = ""

    # merging vehicle alternatives to choosers
    choosers = tours_merged.to_frame().reset_index()
    choosers = pd.merge(choosers, vehicles_wide, how="left", on="household_id")
    choosers.set_index("tour_id", inplace=True)

    # ----- setup skim keys
    skims = get_skim_dict(network_los, choosers)
    locals_dict.update(skims)

    # ------ preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:
        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    logger.info("Running %s with %d tours", trace_label, len(choosers))

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(choosers)

    tours = tours.to_frame()

    # ------ running for each occupancy level selected
    tours_veh_occup_cols = []
    for occup in model_settings.get("OCCUPANCY_LEVELS", [1]):
        logger.info("Running for occupancy = %d", occup)
        # setting occup for access in spec expressions
        locals_dict.update({"occup": occup})

        choices = simulate.simple_simulate(
            choosers=choosers,
            spec=model_spec,
            nest_spec=nest_spec,
            skims=skims,
            locals_d=locals_dict,
            chunk_size=chunk_size,
            trace_label=trace_label,
            trace_choice_name="vehicle_allocation",
            estimator=estimator,
        )

        # matching alt names to choices
        choices = choices.map(dict(enumerate(alts_from_spec))).to_frame()
        choices.columns = ["alt_choice"]

        # last alternative is the non-household vehicle option
        for alt in alts_from_spec[:-1]:
            choices.loc[choices["alt_choice"] == alt, "choice"] = choosers.loc[
                choices["alt_choice"] == alt, alt
            ]
        choices.loc[
            choices["alt_choice"] == alts_from_spec[-1], "choice"
        ] = alts_from_spec[-1]

        # creating a column for choice of each occupancy level
        tours_veh_occup_col = f"vehicle_occup_{occup}"
        tours[tours_veh_occup_col] = choices["choice"]
        tours_veh_occup_cols.append(tours_veh_occup_col)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "households", "vehicle_allocation"
        )
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    pipeline.replace_table("tours", tours)

    tracing.print_summary(
        "vehicle_allocation", tours[tours_veh_occup_cols], value_counts=True
    )

    annotate_settings = model_settings.get("annotate_tours", None)
    if annotate_settings:
        annotate_vehicle_allocation(model_settings, trace_label)

    if trace_hh_id:
        tracing.trace_df(tours, label="vehicle_allocation", warn_if_empty=True)



# create bat file to run all examples from the examples folder
# copies commented out lines to a bat file for running
# python create_run_all_examples.py > run_all_examples.bat
# ben.stabler@rsginc.com 04/05/21

runnable_line_signature = "  # "  # yes, hacky for now
examples_file_name = "example_manifest.yaml"

example_file = open(examples_file_name, "r")
lines = example_file.readlines()
for line in lines:
    if runnable_line_signature in line:
        runnable_line = line.replace(runnable_line_signature, "").replace("\n", "")
        print(runnable_line)



# ActivitySim
# See full license in LICENSE.txt.

import logging

from activitysim.core import config, inject, los
from activitysim.core.pathbuilder import TransitVirtualPathBuilder

logger = logging.getLogger(__name__)

"""
Read in the omx files and create the skim objects
"""


@inject.injectable(cache=True)
def network_los_preload():

    # when multiprocessing with shared data mp_tasks has to call network_los methods
    # allocate_shared_skim_buffers() and load_shared_data() BEFORE network_los.load_data()
    logger.debug("loading network_los_without_data_loaded injectable")
    nw_los = los.Network_LOS()

    return nw_los


@inject.injectable(cache=True)
def network_los(network_los_preload):

    logger.debug("loading network_los injectable")
    network_los_preload.load_data()
    return network_los_preload


@inject.injectable(cache=True)
def skim_dict(network_los):
    return network_los.get_default_skim_dict()


@inject.injectable()
def log_settings():

    # abm settings to log on startup
    return [
        "households_sample_size",
        "chunk_size",
        "chunk_method",
        "chunk_training_mode",
        "multiprocess",
        "num_processes",
        "resume_after",
        "trace_hh_id",
    ]



# ActivitySim
# See full license in LICENSE.txt.

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from ..util import other_than, quick_loc_df, quick_loc_series, reindex


@pytest.fixture(scope="module")
def people():
    return pd.DataFrame(
        {
            "household": [1, 2, 2, 3, 3, 3, 4, 4, 4, 4],
            "ptype": [1, 2, 1, 3, 1, 2, 3, 2, 2, 1],
        },
        index=["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
    )


def test_other_than(people):
    expected = pd.Series(
        [False, False, True, True, True, False, True, True, True, True],
        index=people.index,
        name="left",
    )

    bools = people["ptype"] == 2
    others = other_than(people["household"], bools)

    pdt.assert_series_equal(others, expected)


def test_reindex():
    s = pd.Series([0.5, 1.0, 1.5], index=[2, 1, 3])
    s2 = pd.Series([1, 2, 3], index=["a", "b", "c"])
    assert list(reindex(s, s2).values) == [1.0, 0.5, 1.5]


def test_quick_loc_df():

    df = pd.DataFrame({"attrib": ["1", "2", "3", "4", "5"]}, index=[1, 2, 3, 4, 5])

    loc_list = np.asanyarray([2, 1, 3, 4, 4, 5, 1])
    attrib_list = [str(i) for i in loc_list]

    assert list(quick_loc_df(loc_list, df, "attrib")) == attrib_list
    assert list(quick_loc_df(loc_list, df, "attrib")) == list(
        df.loc[loc_list]["attrib"]
    )


def test_quick_loc_series():

    series = pd.Series(["1", "2", "3", "4", "5"], index=[1, 2, 3, 4, 5])

    loc_list = np.asanyarray([2, 1, 3, 4, 4, 5, 1])
    attrib_list = [str(i) for i in loc_list]

    assert list(quick_loc_series(loc_list, series)) == attrib_list
    assert list(quick_loc_series(loc_list, series)) == list(series.loc[loc_list])



import numpy as np
import pandas as pd

df = pd.read_csv(f"trip_mode_coefficients_p.csv", comment="#")

alts = list(df.drop(columns="Expression").columns.values.astype(str))
alts_str = "_".join(alts)

df = df.set_index("Expression").unstack()

df = df.reset_index().rename(columns={"level_0": "alts", 0: "value"})

df = df.groupby(["Expression", "value"]).agg(lambda col: "_".join(col)).reset_index()

df["coefficient_name"] = "coef_" + np.where(
    df.alts == alts_str, df["Expression"], df["Expression"] + "_" + df.alts
)

coefficients_df = df


df = pd.read_csv(f"trip_mode_coefficients_p.csv", comment="#")

for alt in alts:
    alt_df = pd.merge(
        df[["Expression", alt]].rename(columns={alt: "value"}),
        coefficients_df[["Expression", "value", "coefficient_name"]],
        left_on=["Expression", "value"],
        right_on=["Expression", "value"],
        how="left",
    )
    df[alt] = alt_df["coefficient_name"]

coefficients_df = coefficients_df[["coefficient_name", "value"]]
coefficients_df.to_csv(f"trip_mode_choice_coefficients.csv", index=False)


df.to_csv(f"trip_mode_choice_coefficients_template.csv", index=False)
